Package: AnVILAz
Title: R / Bioconductor Support for the AnVIL Azure Platform
Version: 1.7.2
Date: 2026-05-26
Authors@R: 
    c(person(
        "Martin", "Morgan", role = c("aut", "ctb"),
        email = "mtmorgan.bioc@gmail.com",
        comment = c(ORCID = "0000-0002-5874-8148")
    ),
    person(
        "Marcel", "Ramos", role = c("aut", "cre"),
        email = "marcel.ramos@sph.cuny.edu",
        comment = c(ORCID = "0000-0002-3242-0582")
    ))
Description: The AnVIL is a cloud computing resource developed in part
    by the National Human Genome Research Institute. The AnVILAz
    package supports end-users and developers using the AnVIL platform
    in the Azure cloud. The package provides a programmatic interface
    to AnVIL resources, including workspaces, notebooks, tables, and
    workflows. The package also provides utilities for managing
    resources, including copying files to and from Azure Blob Storage,
    and creating shared access signatures (SAS) for secure access to
    Azure resources.
License: Artistic-2.0
Encoding: UTF-8
Depends: R (>= 4.5.0)
Imports: AnVILBase, BiocBaseUtils, curl, httr2, jsonlite, methods,
        rjsoncons, tibble, utils
biocViews: Software, Infrastructure, ThirdPartyClient
Roxygen: list(markdown = TRUE)
Suggests: BiocStyle, dplyr, knitr, readr, rmarkdown, tinytest
SystemRequirements: az, azcopy
BugReports: https://github.com/Bioconductor/AnVILAz/issues
URL: https://github.com/Bioconductor/AnVILAz
VignetteBuilder: knitr
Collate: 'AnVILAz-package.R' 'authentication.R' 'azure-class.R'
        'avnotebooks-methods.R' 'avtable-methods.R'
        'avworkflow-methods.R' 'avworkspace-methods.R' 'az-utilities.R'
        'az_copy-helpers.R' 'az_sas_token.R' 'azure-methods.R'
        'has_avworkspace-methods.R' 'resources.R' 'utilities.R'
        'workspace-dev-ops.R' 'workspace_env.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/AnVILAz
git_branch: devel
git_last_commit: 89e3a4b
git_last_commit_date: 2026-05-26
Repository: Bioconductor 3.24
Date/Publication: 2026-06-01
NeedsCompilation: no
Packaged: 2026-06-01 21:14:11 UTC; biocbuild
Author: Martin Morgan [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-5874-8148>),
  Marcel Ramos [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3242-0582>)
Maintainer: Marcel Ramos <marcel.ramos@sph.cuny.edu>

# Generated by roxygen2: do not edit by hand

S3method(as.logical,azcopyStatus)
S3method(print,azcopyStatus)
S3method(summary,azcopyStatus)
export(avworkflow_jobs_inputs)
export(az_exists)
export(az_health_check)
export(az_sas_token)
export(azure)
export(cbas_url)
export(wds_api_version)
export(workspace_data_service_url)
export(workspace_id)
export(workspace_storage_cont_id)
export(workspace_storage_cont_url)
exportClasses(azure)
exportMethods(avbackup)
exportMethods(avcopy)
exportMethods(avlist)
exportMethods(avnotebooks)
exportMethods(avnotebooks_delocalize)
exportMethods(avnotebooks_localize)
exportMethods(avremove)
exportMethods(avrestore)
exportMethods(avstorage)
exportMethods(avtable)
exportMethods(avtable_delete)
exportMethods(avtable_delete_values)
exportMethods(avtable_import)
exportMethods(avtable_import_set)
exportMethods(avtables)
exportMethods(avworkflow_jobs)
exportMethods(avworkspace)
exportMethods(avworkspace_clone)
exportMethods(avworkspace_name)
exportMethods(avworkspace_namespace)
exportMethods(avworkspaces)
exportMethods(has_avworkspace)
importClassesFrom(AnVILBase,Platform)
importFrom(AnVILBase,avbackup)
importFrom(AnVILBase,avcopy)
importFrom(AnVILBase,avlist)
importFrom(AnVILBase,avnotebooks)
importFrom(AnVILBase,avnotebooks_delocalize)
importFrom(AnVILBase,avnotebooks_localize)
importFrom(AnVILBase,avremove)
importFrom(AnVILBase,avrestore)
importFrom(AnVILBase,avstorage)
importFrom(AnVILBase,avtable)
importFrom(AnVILBase,avtable_delete)
importFrom(AnVILBase,avtable_delete_values)
importFrom(AnVILBase,avtable_import)
importFrom(AnVILBase,avtable_import_set)
importFrom(AnVILBase,avtables)
importFrom(AnVILBase,avworkflow_jobs)
importFrom(AnVILBase,avworkspace)
importFrom(AnVILBase,avworkspace_clone)
importFrom(AnVILBase,avworkspace_name)
importFrom(AnVILBase,avworkspace_namespace)
importFrom(AnVILBase,avworkspaces)
importFrom(AnVILBase,cloud_platform)
importFrom(AnVILBase,has_avworkspace)
importFrom(BiocBaseUtils,isCharacter)
importFrom(BiocBaseUtils,isScalarCharacter)
importFrom(BiocBaseUtils,isScalarLogical)
importFrom(httr2,req_auth_bearer_token)
importFrom(httr2,req_body_json)
importFrom(httr2,req_body_multipart)
importFrom(httr2,req_cache)
importFrom(httr2,req_headers)
importFrom(httr2,req_method)
importFrom(httr2,req_perform)
importFrom(httr2,req_template)
importFrom(httr2,req_url_query)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_is_error)
importFrom(jsonlite,fromJSON)
importFrom(methods,new)
importFrom(rjsoncons,jmespath)
importFrom(utils,URLencode)
importFrom(utils,stack)
importFrom(utils,write.table)

# AnVILAz

<!-- badges: start -->
<!-- badges: end -->

The AnVIL is a cloud computing resource developed in part
    by the National Human Genome Research Institute. The AnVILAz
    package supports end-users and developers using the AnVIL platform
    in the Azure cloud.

## Installation

You can install the development version of AnVILAz like so:

``` r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager", repos = "https://cran.r-project.org")
BiocManager::install("Bioconductor/AnVILAz")
```

## Example

This is a basic example which shows you how to solve a common problem:

``` r
library(AnVILAz)
```

.onLoad <- function(libname, pkgname) {
    evars <- list(
        workspaceId = workspace_id(),
        wdsApiVersion = wds_api_version(),
        workspaceName = .workspace_name(),
        resourceId = workspace_storage_cont_id(),
        wscu = workspace_storage_cont_url()
    )
    for (k in names(evars)) {
        .avcache$set(k, evars[[k]])
    }
}

#' @name azure-methods
#'
#' @title A number of methods compatible with the Azure platform class.
#'
#' @param source `character(1)` A relative file path corresponding to
#'   either the remote (`az_copy_from_storage`) or local (`az_copy_to_storage`)
#'   file location. Remote locations should be relative to the base directory in
#'   the Azure Storage Container e.g., `analyses/jupyter.log`.
#'
#' @param destination `character(1)` A relative file path corresponding to
#'   either the remote (`az_copy_to_storage`) or local (`az_copy_from_storage`)
#'   file location. Remote locations should be relative to the base directory in
#'   the Azure Storage Container. When not specified, it will default to the
#'   base directory of the remote location. The `to` path can be a folder path
#'   but must end in a forward slash (`/`). If the `to` path points to a
#'   non-existent directory, it will be created.
#'
#' @param recursive `logical(1)` Whether to recursively move or remove files in
#'   a directory. Only applies to `avremove`, `avbackup`, and `avrestore`.
#'   Default is `TRUE` for `backup` and `restore` operations and `FALSE` for
#'   `avremove`.
#'
#' @param dry `logical(1)` Whether to perform a dry run i.e., add the
#'   `--dry-run` flag to the command.
#'
#' @param platform `azure()` The cloud platform class to dispatch on as given by
#'   [AnVILBase::cloud_platform]. Typically not set manually as
#'   `cloud_platform()` returns the `"azure"` class for Azure workspaces on
#'   AnVIL.
#'
#' @param ... Additional arguments passed to lower level functions (not used)
#'
#' @details  The `recursive` argument for `avbackup` and `avrestore` is set to
#'   `TRUE` by default and `FALSE` for `avremove`. Note that wildcards are not
#'   supported for local or remote paths.
#'
#' @return
#' * `avlist` - a `tibble` of files and metadata
#' * `avcopy` - called for the side effect of copying a file __to__ or __from__
#'   the Azure Storage Container depending on the `source` and `destination`
#'   inputs
#' * `avremove` - called for the side effect of removing a file or folder
#' * `avbackup` - called for the side effect of copying a directory __to__
#'   the Azure Storage Container
#' * `avrestore` - called for the side effect of copying a directory
#'   __from__ the Azure Storage Container
#' * `avstorage` - a URL string of the Azure Storage Container location
#' * `avworkspaces` - a tibble of workspaces on AnVIL
#' * `avtable_import` - a response list indicating successful upload
#' * `avtable_delete_values` - when successful, a `NULL` value
#'
#' @examples
#' if (interactive()) {
#'
#'   avlist()
#'
#'   ## local -> remote
#'   ## using general interface avcopy
#'   avcopy("jupyter.log", "analyses/jupyter.log")
#'
#'   ## upload a directory
#'   avbackup("./test/", "analyses/test/")
#'
#'   ## using general interface az_copy
#'   avcopy("analyses/jupyter.log", "./jupyter.log")
#'
#'   ## download a directory
#'   avrestore("analyses/test/", "./test/")
#'
#'   avremove("analyses/jupyter.log")
#'
#' }
NULL

# avcopy ------------------------------------------------------------------

#' @describeIn azure-methods a generalized interface for either
#'   `az_copy_from_storage` or `az_copy_to_storage`; deduced from the `source`
#'   and `destination` inputs
#'
#' @importFrom AnVILBase avcopy
#' @exportMethod avcopy
setMethod("avcopy", signature = c(platform = "azure"), definition =
    function(source, destination, dry = TRUE, ..., platform = cloud_platform())
    {
        stopifnot(
            isScalarCharacter(source)
        )
        if (.is_remote_path(source))
            az_copy_from_storage(from = source, to = destination, dry = dry)
        else if (
            file.exists(source) && (
                .is_remote_path(destination) || missing(destination)
            )
        )
            az_copy_to_storage(from = source, to = destination, dry = dry)
        else
            stop("Invalid source or destination path")
    }
)

# avlist ------------------------------------------------------------------

#' @describeIn azure-methods list all the files in the Azure Storage Container
#'
#' @importFrom AnVILBase avlist
#' @exportMethod avlist
setMethod("avlist", signature = c(platform = "azure"), definition =
    function(..., platform = cloud_platform()) {
        path <- az_sas_token()[["url"]]
        args <- c("list", shQuote(path))
        output <- .az_do("azcopy", args = args)
        files <- strsplit(output, "; ")
        files <- lapply(
            files,
            function(x) gsub("INFO:\\s+|Content\\sLength:\\s+", "", x)
        )
        res <- do.call(rbind.data.frame, files)
        names(res) <- c("INFO", "Content.Length")
        tibble::as_tibble(res)
    }
)

# avremove ----------------------------------------------------------------

#' @describeIn azure-methods remove a file or directory from the Azure Storage
#'
#' @importFrom AnVILBase avremove
#' @exportMethod avremove
setMethod("avremove", signature = c(platform = "azure"), definition =
    function(source, recursive = FALSE, ..., platform = cloud_platform()) {
        stopifnot(
            isScalarCharacter(source)
        )
        .validate_blob(source)
        wscu <- .avcache$get("wscu")
        sas_cred <- az_sas_token()
        token_slug <- sas_cred[["token"]]
        path <- paste0(wscu, "/", source, "?")
        path <- shQuote(paste0(path, token_slug))
        recurse <- paste0("--recursive=", tolower(recursive))
        args <- c("rm", path, recurse)
        results <- .az_do("azcopy", args = args)
        .azcopyStatus(results)
    }
)

# avbackup ----------------------------------------------------------------

#' @describeIn azure-methods copy a directory from the workspace environment to
#'   the Azure Storage Container
#'
#' @importFrom AnVILBase avbackup
#' @exportMethod avbackup
setMethod("avbackup", signature = c(platform = "azure"), definition =
    function(
        source, destination, recursive = TRUE, ...,
        platform = cloud_platform()
    ) {
        if (endsWith(source, "*"))
            stop("Source path cannot end with a wildcard")

        stopifnot(
            isScalarCharacter(source), dir.exists(source),
            is.logical(recursive)
        )

        source <- gsub("\\/$", "", source)
        source <- normalizePath(source)

        wscu <- .avcache$get("wscu")
        sas_cred <- az_sas_token()
        token <- sas_cred[["token"]]
        path <- sas_cred[["url"]]
        if (!missing(destination)) {
            destination <- gsub("\\/$", "", destination)
            path <- paste0(wscu, "/", destination, "?")
            path <- paste0(path, token)
        }

        .az_copy(
            from = shQuote(source), to = shQuote(path),
            paste0("--recursive=", tolower(recursive))
        )
    }
)

# avrestore ---------------------------------------------------------------

#' @describeIn azure-methods copy a file or directory from the Azure Storage
#'   Container to the workspace environment
#'
#' @importFrom AnVILBase avrestore
#' @exportMethod avrestore
setMethod("avrestore", signature = c(platform = "azure"), definition =
    function(
        source, destination, recursive = TRUE, ...,
        platform = cloud_platform()
    ) {
        stopifnot(
            isScalarCharacter(source),
            isScalarCharacter(destination)
        )
        if (!dir.exists(destination))
            dir.create(destination, recursive = TRUE)

        source <- gsub("\\/$", "", source)

        sas_cred <- az_sas_token()
        token <- sas_cred[["token"]]
        wscu <- .avcache$get("wscu")
        path <- paste0(wscu, "/", source, "?")
        path <- paste0(path, token)

        .az_copy(
            from = shQuote(path), to = .az_shQuote(destination),
            paste0("--recursive=", tolower(recursive))
        )
    }
)

# avstorage ---------------------------------------------------------------

#' @describeIn azure-methods The base URI string used to move data to and from
#'   the Azure Storage Container
#'
#' @importFrom AnVILBase avstorage
#' @exportMethod avstorage
setMethod("avstorage", signature = c(platform = "azure"), definition =
    function(..., platform = cloud_platform()) {
        .avcache$get("wscu")
    }
)

#' @name has_avworkspace-methods
#'
#' @title Helper to check if the current environment is within an Azure
#'   workspace
#'
#' @description `has_avworkspace()` checks that the AnVIL environment is set up
#'   to work with Azure. If `strict = TRUE`, it also checks that the workspace
#'   name is set.
#'
#' @inheritParams AnVILBase::has_avworkspace
#'
#' @return `logical(1)` `TRUE` if the AnVIL environment is set up properly to
#'   interact with Azure, otherwise `FALSE`.
#'
#' @examples
#' has_avworkspace(platform = azure())
#'
NULL

#' @describeIn has_avworkspace-methods Check if the AnVIL environment is set up
#'
#' @importFrom AnVILBase has_avworkspace
#'
#' @exportMethod has_avworkspace
setMethod("has_avworkspace", signature = c(platform = "azure"), definition =
    function(strict = FALSE, ..., platform = cloud_platform()) {
        az_exists() &&
            identical(AnVILBase::avplatform_namespace(), "AnVILAz") &&
            (!strict || nzchar(avworkspace_name()))
    }
)

#' @title Generate an Azure authentication token with the `az` command line
#'   utility
#'
#' @description This function generates an Azure authentication token with the
#'   `az` command line utility. The token is used to authenticate with the Azure
#'   services. This function is called internally by the `az_*` functions. It
#'   is not meant to be called directly.
#'
#' @return A character string containing the authentication token with a
#'   "Bearer" prefix.
#'
#' @keywords internal
az_token <- function() {
    system2("az", c("login",  "--identity", "--allow-no-subscriptions"))
    cli_token <- system2("az", c("account", "get-access-token"), stdout = TRUE)
    restkn <- jsonlite::fromJSON(cli_token)
    restkn[["accessToken"]]
}

#' @name avtable-methods
#'
#' @aliases avtable avtables avtable_import avtable_import_set avtable_delete
#' @aliases avtable_delete_values
#'
#' @title AnVIL Azure table ("type") methods
#'
#' @description Methods for working with AnVIL Azure tables. These are referred
#'   to as "types" in the AnVIL Workspace Data Service (WDS) API.
#'
#' @details `avtable_import_set()` creates new rows in a table `<origin>_set`.
#'   One row will be created for each distinct value in the column identified by
#'   `set`. Each row entry has a corresponding column `<origin>` linking to one
#'   or more rows in the `<origin>` table, as given in the `member` column. The
#'   operation is somewhat like `split(member, set)`.
#'
#' @include azure-class.R
#'
#' @inheritParams azure-methods
#'
#' @examples
#' if (interactive()) {
#'   library(dplyr)
#'   mtcars_tbl <-
#'       mtcars |>
#'       as_tibble(rownames = "model_id") |>
#'       mutate(model_id = gsub(" ", "-", model_id))
#'
#'   avtable_import(
#'     mtcars_tbl,
#'     table = "testData",
#'     entity = "model_id"
#'   )
#'
#'   avtable("testData")
#'
#'   avtable("testData") |>                  # new 'testData_set' table
#'         avtable_import_set("testData", "cyl", "model_id")
#'
#'   avtable_delete("testData_set")
#'
#'   avtable_delete_values("testData", "Mazda-RX4")
#' }
NULL


# avtable -----------------------------------------------------------------

#' @describeIn avtable-methods List the contents of a particular table / type
#'
#' @param table `character(1)` The name of the table / type
#'
#' @return `avtable`: a `tibble()` corresponding to the data with the name as
#'   given by `table`
#'
#' @importFrom AnVILBase avtable
#' @importFrom BiocBaseUtils isScalarCharacter
#' @exportMethod avtable
setMethod("avtable", signature = c(platform = "azure"), definition =
    function(table, ..., platform = cloud_platform()) {
        stopifnot(isScalarCharacter(table))
        download_tsv(
            type = table
        )
    }
)

# avtables ----------------------------------------------------------------

#' @describeIn avtable-methods List the available tables / types
#'
#' @return `avtables`: a `tibble()` with columns `table`, `count`, and
#'   `colnames` corresponding to the tables / types available in the current
#'   workspace
#'
#' @importFrom AnVILBase avtables
#' @importFrom rjsoncons jmespath
#' @importFrom jsonlite fromJSON
#' @exportMethod avtables
setMethod("avtables", signature = c(platform = "azure"), definition =
    function(
        ..., platform = cloud_platform()
    ) {
        instanceid <- .avcache$get("workspaceId")
        v <- .avcache$get("wdsApiVersion")

        base_uri <- workspace_data_service_url()
        resp <- request(base_uri) |>
            req_template(
                "/{instanceid}/types/{v}",
                instanceid = instanceid,
                v = v
            ) |>
            req_auth_bearer_token(az_token()) |>
            req_headers("Accept" = "application/json") |>
            req_perform() |>
            resp_body_string()
        tibble::tibble(
            table = jmespath(resp, "[*].name", as = "R"),
            count = jmespath(resp, "[*].count", as = "R"),
            colnames = vapply(
                jmespath(resp, "[*].attributes[*].name", as = "R"),
                function(cnames) paste(cnames, collapse = ", "),
                character(1L)
            )
        )
    }
)

# avtable_import ----------------------------------------------------------

#' @describeIn avtable-methods Upload a dataset to the DATA tab
#'
#' @param entity The entity name, i.e., the name of the column in the table that
#'   provides the keys for the data (a.k.a. `primaryKey`). By default, the first
#'   column in the table. The keys cannot contain special characters or spaces.
#'
#' @return `avtable_import()`: called for the side effect of uploading the data
#'   to the DATA tab
#'
#' @importFrom AnVILBase avtable_import
#' @importFrom utils write.table
#' @exportMethod avtable_import
setMethod("avtable_import", signature = c(platform = "azure"), definition =
    function(
        .data, table, entity = names(.data)[[1L]],
        ..., platform = cloud_platform()
    ) {
        stopifnot(
            is.data.frame(.data), isScalarCharacter(entity)
        )
        if (missing(table))
            table <- deparse(substitute(.data))

        temptsv <- tempfile(fileext = ".tsv")
        write.table(.data, file = temptsv, sep = "\t", row.names = FALSE)
        on.exit(file.remove(temptsv))

        upload_tsv(
            tsv_file = temptsv, type = table, primaryKey = entity
        )
    }
)

# avtable_import_set ------------------------------------------------------

#' @describeIn avtable-methods Create a grouping table from an origin dataset
#'
#' @param .data `tibble()` The dataset chiefly from the `avtable()` operation
#'
#' @param origin `character(1)` name of the type (entity table) used to create
#'   the set e.g "sample", "participant", etc.
#'
#' @param set `character(1)` column name of `.data` identifying the set(s) to be
#'   created, i.e., the grouping variable.
#'
#' @param member `character(1)` column name of `.data` identifying the member(s)
#'   of the set(s) or groups. The values in this column may repeat if
#'   an ID is in more than one set.
#'
#' @return `avtable_import_set()`: a `character(1)` name of the imported
#'   tibble.
#'
#' @importFrom AnVILBase avtable_import_set
#' @importFrom BiocBaseUtils isScalarCharacter
#' @importFrom utils URLencode stack
#' @exportMethod avtable_import_set
setMethod("avtable_import_set", signature = c(platform = "azure"),
    definition = function(
        .data, origin, set = names(.data)[[1]], member = names(.data)[[2]],
        ..., platform = cloud_platform()
    ) {
        stopifnot(
            is.data.frame(.data),
            isScalarCharacter(origin),
            isScalarCharacter(set),
            isScalarCharacter(member),
            set %in% names(.data),
            !identical(set, member), member %in% names(.data)
        )
        origin <- URLencode(origin)
        .data <- .data[, c(set, member)]
        .data <- rev(stack(
            lapply(
                split(.data, .data[[1L]]),
                function(x) paste(x[[2L]], collapse = ", ")
            )
        ))
        names(.data)[[1L]] <- paste0(origin, "_set_id")
        names(.data)[[2L]] <- origin

        table_type <- paste0(origin, "_set")
        avtable_import(
            .data, table_type, entity = names(.data)[[1L]]
        )
        table_type
    }
)


# avtable_delete ----------------------------------------------------------

#' @describeIn avtable-methods Delete a table / type
#' @importFrom AnVILBase avtable_delete
#'
#' @return `avtable_delete`: a `logical(1)` indicating success or failure
#'
#' @exportMethod avtable_delete
setMethod("avtable_delete", signature = c(platform = "azure"), definition =
    function(table, ..., platform = cloud_platform()) {
        stopifnot(isScalarCharacter(table))
        delete_type(type = table)
    }
)

# avtable_delete_values ---------------------------------------------------

#' @describeIn avtable-methods Delete rows from a table / type
#'
#' @param values `character()` vector of `primaryKey` values corresponding to
#'   rows to be deleted
#'
#' @return `avtable_delete_values()`: a `logical(1)` vector indicating success
#'   or failure for each value in `values`
#'
#' @importFrom AnVILBase avtable_delete_values
#' @exportMethod avtable_delete_values
setMethod("avtable_delete_values", signature = c(platform = "azure"),
    definition = function(table, values, ..., platform = cloud_platform()) {
        stopifnot(
            isScalarCharacter(table), isCharacter(values)
        )

        tsv <- download_tsv(type = table)
        allids <- tsv[[1L]]
        stopifnot(all(values %in% allids))

        names(values) <- values
        vapply(values, function(val) {
            delete_type_id(
                type = table,
                id = val
            )
        }, logical(1L))
    }
)

## ----setup, include = FALSE----------------------------------------------
az_ok <- AnVILAz::has_avworkspace(strict = TRUE, platform = AnVILAz::azure())
knitr::opts_chunk$set(
    collapse = TRUE,
## Related to https://stat.ethz.ch/pipermail/bioc-devel/2020-April/016656.html
    crop = NULL,
    eval = az_ok
)
options(width = 75)

## ----library, message = FALSE, eval = TRUE, cache = FALSE----------------
library(AnVILAz)

## ----avworkspaces, eval=az_ok--------------------------------------------
# avworkspaces()

## ----avworkspace, eval=az_ok---------------------------------------------
# avworkspace()

## ----avworkspace_set, eval=az_ok-----------------------------------------
# avworkspace("my-namespace/my-workspace")
# avworkspace()

## ----avworkspace_namespace, eval=az_ok-----------------------------------
# avworkspace_namespace()

## ----avworkspace_name, eval=az_ok----------------------------------------
# avworkspace_name()

## ----avworkspace_clone, eval=FALSE---------------------------------------
# avworkspace_clone(
#     namespace = "my-namespace",
#     name = "my-workspace",
#     to_namespace = "my-new-namespace",
#     to_name = "my-workspace-clone"
# )
# avworkspaces()

## ----avnotebooks, eval=az_ok---------------------------------------------
# avnotebooks()

## ----avnotebooks_localize, eval=FALSE------------------------------------
# avnotebooks_localize(destination = "./analyses")

## ----avnotebooks_delocalize, eval=FALSE----------------------------------
# avnotebooks_delocalize(source = "./")

## ----avworkflows_jobs, eval=az_ok----------------------------------------
# avworkflows_jobs()

## ----avworkflows_jobs_inputs, eval=az_ok---------------------------------
# avworkflows_jobs_inputs()

## ----sessionInfo, eval=TRUE----------------------------------------------
sessionInfo()

.DSDE_PROD_URL <- "https://workspace.dsde-prod.broadinstitute.org"

query_resources <- function(FUN = resp_body_json) {
    request(.DSDE_PROD_URL) |>
        req_template(
            "/api/workspaces/v1/{workspaceId}/resources",
            workspaceId = .avcache$get("workspaceId")
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_url_query(stewardship = "CONTROLLED", limit = 1000) |>
        req_perform() |>
        FUN()
}

query_userName <- function() {
    userName <- query_resources(httr2::resp_body_string) |>
        rjsoncons::jmespath(
            paste0(
                "resources[?contains(metadata.resourceType,'AZURE_VM')].",
                "metadata.controlledResourceMetadata.privateResourceUser.",
                "userName"
            )
        ) |> jsonlite::fromJSON()
}

#' @docType class
#' @title Azure platform class
#' @description This class represents the Azure platform.
#' @name azure-class
#'
#' @importClassesFrom AnVILBase Platform
#' @importFrom methods new
#' @importFrom AnVILBase cloud_platform
#'
#' @returns An object of class `azure`.
#'
#' @examples
#' showClass("azure")
#'
#' @exportClass azure
.azure <- setClass("azure", contains = "Platform")

#' @rdname azure-class
#'
#' @examples
#' az <- azure()
#' az
#' showClass(class(az))
#' @export
azure <- function() {
    .azure()
}

#' @name workspace-env
#'
#' @aliases workspace_id workspace_storage_container_id
#' @aliases workspace_storage_container_url workspace_data_service_url
#'
#' @title Access Terra on Azure workspace session variables
#'
#' @description A group of functions that return environment variables in the
#'   Terra Azure workspace. The Workspace Data Service URL sends out an API
#'   `GET` request to obtain the data services URL for uploading data to the
#'   workspace "DATA" tab.
#'
#' @return
#' * `workspace_id` - A UUID string referring to "workspaceId" or "workspaceid"
#'   in API calls
#' * `workspace_storage_cont_id` - A UUID string identifiying the resource
#'   storage container owned by the user account, a.k.a. "resourceId"
#' * `workspace_storage_cont_url` - The base URI string used to move data
#'   to and from the Azure Storage Container
#' * `wds_api_version` - The version of the Workspace Data Service API, defaults
#'   to "v0.2"
#' * `workspace_data_service_url` - The base URI string used to move data to
#'   to and from the workspace "DATA" tab
#' * `cbas_url` - The base URI string used to query the workflow submission
#'   history
#'
#' @examples
#' workspace_id()
#' workspace_storage_cont_id()
#' workspace_storage_cont_url()
#' if (interactive()) {
#'   workspace_data_service_url()
#'   cbas_url()
#' }
#' @export
workspace_id <- function() {
    opt <- Sys.getenv("WORKSPACE_ID", "")
    getOption("AnVILAz.workspace_id", opt)
}

.workspace_name <- function() {
    opt <- Sys.getenv("WORKSPACE_NAME", "")
    getOption("AnVILAz.workspace_name", opt)
}

#' @rdname workspace-env
#' @export
workspace_storage_cont_id <- function() {
    opt <- Sys.getenv("WORKSPACE_STORAGE_CONTAINER_ID", "")
    getOption("AnVILAz.workspace_storage_cont_id", opt)
}

#' @rdname workspace-env
#' @export
workspace_storage_cont_url <- function() {
    opt <- Sys.getenv("WORKSPACE_STORAGE_CONTAINER_URL", "")
    getOption("AnVILAz.workspace_storage_cont_url", opt)
}

#' @rdname workspace-env
#' @export
wds_api_version <- function() {
    opt <- Sys.getenv("WDS_API_VERSION", "v0.2")
    getOption("AnVILAz.wds_api_version", opt)
}

## from terra-workspace-data-service/docs/WDS Python Client.md

#' @rdname workspace-env
#' @export
workspace_data_service_url <- function() {
    .leo_apps("wds")
}

#' @rdname workspace-env
#' @export
cbas_url <- function() {
    .leo_apps("cbas")
}

.list_apps <- function() {
    token <- try({az_token()}, silent = TRUE)
    if (inherits(token, "try-error"))
        return("[]")
    request(.LEONARDO_URL) |>
        req_template(
            "/api/apps/v2/{workspaceId}",
            workspaceId = .avcache$get("workspaceId")
        ) |>
        req_auth_bearer_token(token) |>
        req_url_query(includeDeleted = "false") |>
        req_perform() |>
        resp_body_string()
}

.apps_running <- function() {
    status <- .list_apps() |>
        jsonlite::fromJSON()
    if (!length(status))
        return(NULL)
    status <- status[, c("status", "appType")]
    res <- status[["status"]] == "RUNNING"
    names(res) <- status[["appType"]]
    res
}

.leo_apps <- function(app) {
    url_resp <- .list_apps()
    res_url <- rjsoncons::jmespath(
        url_resp,
        paste0("[*].proxyUrls.", app)
    )
    jsonlite::fromJSON(res_url)
}

# .avcache ----------------------------------------------------------------

# Cache for workspace environment variables
.avcache <- local({
    hash <- new.env(parent = emptyenv())
    list(
        get = function(key) {
            hash[[key]]
        },
        set = function(key, value) {
            if (!is.null(value))
                hash[[key]] <- value
            hash[[key]]
        },
        keys = function() {
            names(hash)
        },
        reset = function() {
            rm(list = ls(hash), envir = hash)
        }
    )
})

#' @name avworkflow-methods
#'
#' @aliases avworkflow_jobs
#'
#' @title Azure Workflow methods
#'
#' @description `avworkflow_jobs()` reports the status of workflow executions
#'    in the current workspace.
#'
#' @details The `avworkflow_jobs_inputs()` function returns the input parameters
#'   for the workflow jobs as a `tibble`.
#'
#' @return `avworkflow_jobs()` returns a tibble with the status of the
#'   jobs in the current workspace.
#'
#' @inheritParams azure-methods
#' @inheritParams avnotebooks-methods
#'
#' @include azure-class.R
#'
#' @examples
#' if (has_avworkspace(strict = TRUE, platform = azure()))
#'     ## from within AnVIL
#'     avworkflow_jobs()
#'
NULL

.jmespath_template <-
    function(x, value, base = "run_sets[*]")
{
    rjsoncons::jmespath(x, paste(base, value, sep = "."), as = "R")
}

.CBAS_VALUES <- c(
    "run_set_id", "run_set_name", "run_set_description", "state",
    "submission_timestamp", "last_modified_timestamp", "run_count",
    "error_count", "user_id"
)

#' @importFrom rjsoncons jmespath
.avworkflow_job <-
    function(x)
{
    vals <- structure(.CBAS_VALUES, .Names = .CBAS_VALUES)
    listvals <- lapply(vals, .jmespath_template, x = x)
    do.call(data.frame, listvals) |>
        tibble::as_tibble()
}

#  avworkflow_jobs ----------------------------------------------------------

#' @describeIn avworkflow-methods List the status of workflow jobs
#'
#' @importFrom AnVILBase avworkflow_jobs
#' @exportMethod avworkflow_jobs
setMethod("avworkflow_jobs", signature = c(platform = "azure"),
    definition = function(
        ...,
        platform = cloud_platform()
    ) {
        qrs <- .avrun_sets()
        .avworkflow_job(qrs)
    }
)

.avrun_sets <- function() {
    request(cbas_url()) |>
        req_template(
            "/api/batch/v1/run_sets"
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_perform() |>
        resp_body_string()
}

#' @rdname avworkflow-methods
#'
#' @export
avworkflow_jobs_inputs <- function() {
    qrs <- .avrun_sets()
    input_tbls <- lapply(
        .jmespath_template(qrs, "input_definition"),
        jsonlite::fromJSON
    )
    names(input_tbls) <- .jmespath_template(qrs, "run_set_id")
    input_tbls
}

#' @name workspace-dev-ops
#'
#' @title Functions to work with workspace data for developers
#'
#' @description These group of functions will allow you to manipulate tables in
#'   the "DATA" tab. Example operations include moving a flat Tab-Separated
#'   Values (TSV) file into the workspace, deleting records, deleting tables,
#'   adding a single row, retrieving a single row, and retrieving the data
#'   table. Note that the API used refers to tables as `types`.
#'
#' @details These functions use the Workspace Data Services (WDS) API. Current
#'   operations that affect the "DATA" tab include:
#'   * `upload_tsv` - a `POST` request using a TSV file that populates the data
#'   * `download_tsv` - a `GET` request with the data name (`type` argument) in
#'     `upload_tsv` to represent the data locally as a `tibble`
#'   * `delete_type_id` - a `DELETE` request to remove a record or row from
#'     `type`
#'   * `add_type_id` - a `PUT` request to add a single row to an existing table
#'     (`type`)
#'   * `get_type_id` - a `GET` request to retrieve a single row from an existing
#'     table (`type`)
#'   * `delete_type` - a `DELETE` request to remove then entire data set
#'     (`type`)
#'
#' @param tsv_file `character(1)` A path to a tab-separated values file
#'
#' @param row `tibble()` or `data.frame()` A single row to add to an existing
#'   table. The row must have the same column names as the table. The
#'   `primaryKey` column must be unique.
#'
#' @param type `character(1)` A nickname for the uploaded dataset important for
#'   retreival. By default, the file name will be used.
#'
#' @param id `character(1)` The value in the `primaryKey` column that
#'   indicates the row to be removed.
#'
#' @param primaryKey `character(1)` The optional column name to uniquely
#'   identify a record. By default, the first column is used as the primary
#'   key and all values in the column must be unique.
#'
#' @return
#'   * `upload_tsv` - A response list indicating successful upload
#'   * `download_tsv` - A `tibble` corresponding to the data labeled with `type`
#'   * `delete_type_id`; `delete_type` - When successful, a `NULL` value
#'
#' @keywords internal
#'
#' @examples
#' if (interactive()) {
#'   library(dplyr)
#'   type <- "model"
#'   mtcars_tbl <-
#'       mtcars |>
#'       as_tibble(rownames = "model_id") |>
#'       mutate(model_id = gsub(" ", "-", model_id))
#'
#'   tsv_file <- tempfile()
#'   readr::write_tsv(mtcars_tbl, tsv_file)
#'   upload_tsv(
#'     tsv_file = tsv_file,
#'     type = "testData",
#'     primaryKey = "model_id"
#'   )
#'
#'   download_tsv("testData")
#'
#'   ## create an example single row tibble for add_type_id
#'   datsun <- filter(mtcars_tbl, model_id == "Datsun-710")
#'   ## change the model_id to be unique
#'   datsun[["model_id"]] <- "Datsun-512"
#'
#'   add_type_id(row = datsun, type = "testData")
#'
#'   get_type_id("testData", "Datsun-512")
#'
#'   delete_type_id("testData", "Datsun-512")
#'
#'   delete_type("testData")
#' }
upload_tsv <- function(
    tsv_file,
    type = tools::file_path_sans_ext(basename(tsv_file)),
    primaryKey = NULL
) {
    base_uri <- workspace_data_service_url()
    request(base_uri) |>
        req_template(
            "/{instanceid}/tsv/{v}/{type}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            type = type
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_url_query(primaryKey = primaryKey) |>
        req_body_multipart(
            records = curl::form_file(
                tsv_file, type = "text/tab-separated-values"
            )
        ) |>
        req_perform() |>
        resp_body_json()
}

#' @rdname workspace-dev-ops
download_tsv <- function(type) {
    opt <- options(readr.show_col_types = FALSE)
    on.exit(options(opt))

    if (!requireNamespace("readr", quietly = TRUE))
        stop("Install the 'readr' package to import TSV files")

    base_uri <- workspace_data_service_url()
    request(base_uri) |>
        req_template(
            "/{instanceid}/tsv/{v}/{type}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            type = type
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_perform() |>
        resp_body_string() |>
        readr::read_tsv(show_col_types = FALSE)
}

#' @rdname workspace-dev-ops
delete_type_id <- function(type, id) {
    opt <- options(readr.show_col_types = FALSE)
    on.exit(options(opt))

    base_uri <- workspace_data_service_url()
    response <- request(base_uri) |>
        req_template(
            "/{instanceid}/records/{v}/{type}/{id}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            type = type,
            id = id
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_method("DELETE") |>
        req_perform() |>
        resp_is_error()

    !response
}

#' @rdname workspace-dev-ops
add_type_id <- function(row, type, id = row[[1L]]) {
    stopifnot(identical(nrow(row), 1L))

    opt <- options(readr.show_col_types = FALSE)
    on.exit(options(opt))

    tsv <- download_tsv(type = type)
    primaryKey <- names(tsv)[[1L]]
    allids <- tsv[[primaryKey]]

    if (!primaryKey %in% names(row) && missing(id))
        stop("Primary key not found in row or missing id argument.")
    else if (id %in% allids)
        warning("Replacing record with id: ", id)

    row <- row[, !names(row) %in% primaryKey]

    base_uri <- workspace_data_service_url()
    result <- request(base_uri) |>
        req_template(
            "/{instanceid}/records/{v}/{type}/{id}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            type = type,
            id = id
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_body_json(
            list(attributes = as.list(row))
        ) |>
        req_url_query(primaryKey = primaryKey) |>
        req_method("PUT") |>
        req_perform() |>
        resp_body_json()

    result <- tibble::as_tibble(
        c(key = id, result[["attributes"]])
    )
    names(result)[[1L]] <- primaryKey
    result
}

#' @rdname workspace-dev-ops
get_type_id <- function(type, id) {
    opt <- options(readr.show_col_types = FALSE)
    on.exit(options(opt))

    base_uri <- workspace_data_service_url()
    response <- request(base_uri) |>
        req_template(
            "/{instanceid}/records/{v}/{type}/{id}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            id = id
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_perform() |>
        resp_body_json()

    response[["attributes"]] |>
        tibble::as_tibble()
}

#' @rdname workspace-dev-ops
delete_type <- function(type) {
    opt <- options(readr.show_col_types = FALSE)
    on.exit(options(opt))

    base_uri <- workspace_data_service_url()
    result <- request(base_uri) |>
        req_template(
            "/{instanceid}/types/{v}/{type}",
            instanceid = .avcache$get("workspaceId"),
            v = .avcache$get("wdsApiVersion"),
            type = type
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_method("DELETE") |>
        req_perform() |>
        resp_is_error()

    !result
}

## from AnVILGCP
.az_shQuote <- function(source) {
    ## Expand local paths with ~ or . or .. to full path names.
    ## Needed because we also use shQuote() (to allow for spaces in
    ## file names), and shQuote() would otherwise use paths with ~ or
    ## . in the current working directory.
    source <- normalizePath(source)
    shQuote(source)
}

.is_remote_path <- function(path) {
    startsWith(path, "http") || !file.exists(path)
}

.validate_blob <- function(blob) {
    file_tbl <- avlist()
    allfiles <- file_tbl[["INFO"]]
    is_dir <- endsWith(blob, "/")
    if (is_dir && !any(startsWith(allfiles, blob)))
        stop("Virtual directory not found; check path with `avlist`")

    if (!blob %in% allfiles && !is_dir)
        stop("File not found; check path to blob file with `avlist`")

    TRUE
}

.az_sanitize <- function(x) {
    gsub("(https?://[^? \"]+)\\?[^ \"]+", "\\1?[...]", x)
}

.az_system2 <- function(...) {
    base::system2(...)
}

#' @importFrom BiocBaseUtils isScalarCharacter isCharacter
.az_do <- function(command, args) {
    stopifnot(
        isScalarCharacter(command),
        isCharacter(args, na.ok = FALSE)
    )
    bin <- .az_find(command)
    res <- withCallingHandlers({
        tryCatch({
            .az_system2(bin, args, stdout = TRUE, stderr = TRUE, wait=TRUE)
        }, error = function(err) {
            sani_args <- .az_sanitize(args)
            msg <- paste0(
                "'", command, " ", paste(sani_args, collapse = " "),
                "' failed:\n", "  ", conditionMessage(err)
            )
            stop(msg, call. = FALSE)
        })
    }, warning = function(warn) {
        invokeRestart("muffleWarning")
    })
    if (!is.null(attr(res, "status"))) {
        msg <- paste0(
            "'", command, " ", paste(args, collapse = " "), "' failed:",
            "\n  ", paste(as.vector(res), collapse = "\n    "),
            "\n  exit status: ", attr(res, "status")
        )
        stop(msg, call. = FALSE)
    }
    res
}

.az_find <- function(command) {
    bin <- Sys.which(command)
    if (nzchar(bin))
        normalizePath(bin)
    else
        stop("failed to find '", command, "' binary", call. = FALSE)
}

.az_copy <- function(from, to, ...) {
    args <- c("copy", from, to, ...)
    .az_do("azcopy", args = args)
}

## tinytest helpers

.is_anvil_az <- function() {
    nzchar(Sys.getenv("WORKSPACE_ID"))
}

.is_anvil_gcp <- function() {
    nzchar(Sys.getenv("WORKSPACE_BUCKET"))
}

.exit_if_not_anvilaz <- function() {
    if (!.is_anvil_az())
        tinytest::exit_file("Not running on AnVIL Azure workspace")
}

AnVILAz:::.exit_if_not_anvilaz()

if (!AnVILAz:::.is_anvil_az())
    tinytest::exit_file("Not running on AnVIL Azure workspace")

tinytest::exit_if_not(
    "Not running on AnVIL Azure workspace" = AnVILAz:::.is_anvil_az()
)

## https://github.com/markvanderloo/tinytest/issues/126
## fallback to logical condition for now

if (AnVILAz:::.is_anvil_az()) {

library(BiocBaseUtils)
# avworkspaces ------------------------------------------------------------

aws <- avworkspaces()

expect_true(
    is_tibble(aws)
)

expect_true(
    all(
        c("name", "namespace", "createdBy") %in% colnames(aws)
    )
)

awsns <- avworkspace_namespace()

expect_true(
    isScalarCharacter(awsns)
)

awsn <- avworkspace_name()

expect_true(
    isScalarCharacter(awsn)
)

expect_identical(
    avworkspace(), paste0(awsns, "/", awsn)
)

}

AnVILAz:::.exit_if_not_anvilaz()

if (!AnVILAz:::.is_anvil_az())
    tinytest::exit_file("Not running on AnVIL Azure workspace")

tinytest::exit_if_not(
    "Not running on AnVIL Azure workspace" = AnVILAz:::.is_anvil_az()
)

## https://github.com/markvanderloo/tinytest/issues/126
## fallback to logical condition for now

if (AnVILAz:::.is_anvil_az()) {

library(dplyr)
library(tibble)

type <- "model"
primaryKey <- "model_id"
mtcars_tbl <-
    mtcars |>
    as_tibble(rownames = primaryKey) |>
    mutate(model_id = gsub(" ", "-", model_id))

tsv_file <- tempfile()
readr::write_tsv(mtcars_tbl, tsv_file)

# avtable_import and upload_tsv -------------------------------------------

avtable_import(tsv_file, "testData", primaryKey)

# avtables ----------------------------------------------------------------

avtab <- avtables()
expect_true(
    is_tibble(avtab)
)

expect_identical(
    avtab[["table"]], "testData"
)

expect_identical(
    avtab[["count"]], nrow(mtcars_tbl)
)

expect_identical(
    setequal(avtab[["columns"]], colnames(mtcars_tbl))
)

expect_true(
    setequal(
        strsplit(avtab[["colnames"]], ", ")[[1L]],
        colnames(mtcars_tbl)
    )
)


# avtable and download_tsv ------------------------------------------------

avtab <- avtable("testData")
expect_true(
    is_tibble(avtab)
)
expect_identical(
    colnames(avtab)[1L], primaryKey
)
expect_true(
    setequal(
        colnames(avtab), colnames(mtcars_tbl)
    )
)
expect_identical(
    sort(avtab[[primaryKey]]),
    sort(mtcars_tbl[[primaryKey]])
)
expect_identical(
    dim(avtab),
    dim(mtcars_tbl)
)

# avtable_import_set ------------------------------------------------------

avtable("testData") |>
    avtable_import_set("testData", "cyl", "model_id")

tset <- avtable("testData_set")

expect_identical(
    colnames(tset),
    c("testData_set_id", "testData")
)

expect_identical(
    tset[["testData_set_id"]],
    unique(sort(mtcars_tbl[["cyl"]]))
)

Map(
    function(orig, set) {
        expect_true(
            setequal(orig, set)
        )
    },
    orig = lapply(split(mtcars_tbl, mtcars_tbl[["cyl"]]), `[[`, "model_id"),
    set = unlist(lapply(tset[["testData"]], strsplit, ", "), recursive = FALSE)
)

delete_type("testData")

# avtable_delete_values ---------------------------------------------------

avtable_delete_values("testData_set", "6")

tset <- avtable("testData_set")

expect_identical(
    dim(tset), c(2L, 2L)
)

expect_identical(
    tset[["testData_set_id"]], c(4, 8)
)

delete_type("testData_set")

}

sas_token <- paste0(
    "?sv=2020-08-04&ss=b&srt=sco&sp=rwdlacx&se=2026-12-31T23:59:59Z&",
    "sig=SAS_SIGNATURE"
)

fake_url <-
    paste0("https://mystorage.blob.core.windows.net/container/file", sas_token)

mock_system2 <- function(command, args, stdout, stderr, wait)
    stop("Simulated system2 error: invalid flag used")

orig_fun <- AnVILAz:::.az_system2
unlockBinding(".az_system2", asNamespace("AnVILAz"))
assign(".az_system2", mock_system2, envir = asNamespace("AnVILAz"))

err <- tryCatch(
    {
        AnVILAz:::.az_do("ls", c("--non-existent-flag", fake_url))
    }, error = function(e) conditionMessage(e)
)

expect_true(
    grepl("https://mystorage.blob.core.windows.net/container/file", err),
    info = "Error message should still contain the target URL path for context"
)

expect_false(
    grepl("SAS_SIGNATURE", err),
    info = "Signature was found in the error message!"
)

expect_true(
    grepl("\\?\\[\\.\\.\\.\\]", err),
    info = "The query parameters should be replaced by '[...]'"
)

assign(".az_system2", orig_fun, envir = asNamespace("AnVILAz"))
lockBinding(".az_system2", asNamespace("AnVILAz"))

AnVILAz:::.exit_if_not_anvilaz()

if (!AnVILAz:::.is_anvil_az())
    tinytest::exit_file("Not running on AnVIL Azure workspace")

tinytest::exit_if_not(
    "Not running on AnVIL Azure workspace" = AnVILAz:::.is_anvil_az()
)

## https://github.com/markvanderloo/tinytest/issues/126
## fallback to logical condition for now

if (AnVILAz:::.is_anvil_az()) {

# test avlist -------------------------------------------------------

files <- avlist()
expect_true(
    tibble::is_tibble(files) && is.data.frame(files)
)
expect_identical(
    names(files), c("INFO", "Content.Length")
)

# test az_copy_to and from_storage ----------------------------------------

file.create("test.log")
az_copy_to_storage("test.log", "analyses/test/test.log")
expect_true(
    "analyses/test/test.log" %in% avlist()[["INFO"]]
)
file.remove("test.log")
expect_error(
    az_copy_from_storage("analyses/test/test.log", "./test.log", dry = FALSE)
)
az_copy_from_storage("analyses/test/test.log", "./", dry = FALSE)
expect_true(
    file.exists("test.log")
)

# test avremove ---------------------------------------------------------

avremove("analyses/test/test.log")
expect_false(
    "analyses/test/test.log" %in% avlist()[["INFO"]]
)
expect_false(
    "analyses/test" %in% avlist()[["INFO"]]
)

file.remove("test.log")

# test avbackup and restore -----------------------------------------

dir.create("test")
file.create("test/test.log")
file.create("test/test2.log")

## create remote folder and copy file
expect_error(
    avbackup("./test", "analyses/test_backup/")
)
expect_error(
    avbackup("./test/*", "analyses/test_backup/")
)
expect_true(
    "analyses/test_backup/test/test.log" %in% avlist()[["INFO"]]
)
expect_true(
    "analyses/test_backup/test/test2.log" %in% avlist()[["INFO"]]
)

avremove("analyses/test_backup/", recursive = TRUE)
expect_false(
    "analyses/test_backup/test/test.log" %in% avlist()[["INFO"]]
)
expect_false(
    "analyses/test_backup/test/test2.log" %in% avlist()[["INFO"]]
)

avremove("analyses/test_backup/", recursive = TRUE)

avbackup("./test", "analyses/test_restore")
avrestore(
    "analyses/test_restore/test", "./test_restore"
)
expect_true(
    dir.exists("test_restore/")
)
expect_true(
    file.exists("test_restore/test/test.log")
)

unlink("./test", recursive = TRUE)
unlink("./test_restore", recursive = TRUE)

avrestore("analyses/test_restore", "./test_restore")
expect_true(
    dir.exists("test_restore/test_restore")
)
expect_true(
    file.exists("test_restore/test_restore/test/test.log")
)

unlink("./test_restore", recursive = TRUE)
avremove("analyses/test_restore/", recursive = TRUE)

}

#' @name avnotebooks-methods
#'
#' @aliases avnotebooks avnotebooks_localize avnotebooks_delocalize
#'
#' @title Azure Notebook Management
#'
#' @description `avnotebooks()` lists the notebooks in the current workspace.
#'
#' @param local `logical(1)` notebooks located on the workspace
#'   (`local = FALSE`, default) or runtime / local instance (`local = TRUE`).
#'   When `local = TRUE`, the notebook path is
#'   `<workspace_data_service_url()>/analyses`.
#'
#' @inheritParams azure-methods
#'
#' @return `avnotebooks()` returns a character vector of files located
#'   in the workspace 'analyses/' folder path, or on the local file
#'   system.
#'
#' @include azure-class.R
#'
#' @examples
#' if (has_avworkspace(strict = TRUE, platform = azure()))
#'     avnotebooks()
#'
NULL

# avnotebooks ------------------------------------------------------------

#' @describeIn avnotebooks-methods List the notebooks in the current workspace
#'
#' @importFrom BiocBaseUtils isScalarLogical
#' @importFrom AnVILBase avnotebooks
#' @exportMethod avnotebooks
setMethod("avnotebooks", signature = c(platform = "azure"),
    definition = function(
        local = FALSE,
        ...,
        platform = cloud_platform()
    ) {
        stopifnot(
            isScalarLogical(local)
        )
        if (!local)
            files <- Filter(
                function(x) startsWith(x, "analyses/"), avlist()[["INFO"]]
            )
        else
            files <- list.files()
        isNotebook <- tolower(tools::file_ext(files)) %in% c("ipynb", "rmd")
        files[isNotebook]
    }
)

# avnotebooks_localize ----------------------------------------------------

#' @describeIn avnotebooks-methods Sync notebooks between the Azure Blob Storage
#'   Container and the local runtime
#'
#' @param destination `character(1)` or `missing` file path to the local file
#'   system directory for synchronization. The default location is
#'   `~/<workspace_data_service_url()>/analyses`. Out-of-date local files are
#'   replaced with the workspace version.
#'
#' @importFrom AnVILBase avnotebooks_localize
#' @exportMethod avnotebooks_localize
setMethod("avnotebooks_localize", signature = c(platform = "azure"),
    definition = function(
        destination = "./analyses",
        dry = TRUE,
        ...,
        platform = cloud_platform()
    ) {
        stopifnot(
            missing(destination) || isScalarCharacter(destination),
            isScalarLogical(dry)
        )

        source <- file.path(workspace_data_service_url(), "analyses")
        if (missing(destination)) {
            if (!dry && !dir.exists(destination))
                dir.create(destination, recursive = TRUE)
        }
        az_copy_from_storage(
            from = source, to = destination, recursive = TRUE, dry = dry
        )
    }
)

# avnotebooks_delocalize --------------------------------------------------

#' @describeIn avnotebooks-methods Sync notebooks between the local runtime and
#'   the Azure Blob Storage Container
#'
#' @param source `character(1)` or `missing` file path to the local file system
#'   directory for synchronization. The default location is the home folder.
#'   Out-of-date local files are replaced with the workspace version.
#'
#' @importFrom AnVILBase avnotebooks_delocalize
#' @exportMethod avnotebooks_delocalize
setMethod("avnotebooks_delocalize", signature = c(platform = "azure"),
    definition = function(
        source = "./",
        dry = TRUE,
        ...,
        platform = cloud_platform()
    ) {
        stopifnot(
            missing(source) || isScalarCharacter(source),
            isScalarLogical(dry)
        )

        files <- list.files(path = source)
        isNotebook <- tolower(tools::file_ext(files)) %in% c("ipynb", "rmd")
        notebooks <- file.path(source, files[isNotebook])
        if (!length(notebooks))
            stop("No notebooks found in ", source)

        lapply(
            notebooks,
            function(notebook)
                az_copy_to_storage(
                    from = notebook, to = "analyses/",
                    recursive = FALSE, dry = dry
                )
        )
    }
)

#' Obtain the Shared Access Signature (SAS) for the Azure Storage
#'
#' The function provides a user delegation SAS token for management of
#' resources. Mainly used in other functions to move files to and from the
#' Azure Storage Container
#'
#' @param sasExpirationDuration `numeric(1)` The number of seconds until the SAS
#'   token expires (default: 28,800 seconds)
#'
#' @return A list of two elements named `token` and `url`
#'
#' @examples
#' if (interactive()) {
#'   sas <- az_sas_token()
#'   sas[["token"]]
#'   sas[["url"]]
#' }
#' @export
az_sas_token <- function(sasExpirationDuration = 28800) {
    api_endpoint <- paste0(
        "/api/workspaces/v1/{workspaceId}/resources/",
        "controlled/azure/storageContainer/{resourceId}/getSasToken"
    )
    request(.DSDE_PROD_URL) |>
        req_template(
            api_endpoint,
            workspaceId = .avcache$get("workspaceId"),
            resourceId = .avcache$get("resourceId")
        ) |>
        req_auth_bearer_token(az_token()) |>
        req_url_query(sasExpirationDuration = sasExpirationDuration) |>
        req_method("POST") |>
        req_perform() |>
        resp_body_json()
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom httr2 request
#' @importFrom httr2 req_auth_bearer_token req_headers req_cache req_perform
#' @importFrom httr2 req_method req_template req_body_multipart req_body_json
#' @importFrom httr2 resp_is_error req_url_query
#' @importFrom httr2 resp_body_string resp_body_json
## usethis namespace: end
NULL

#' @name avworkspace-methods
#'
#' @title AnVIL Azure Workspace methods
#'
#' @inheritParams azure-methods
#'
#' @include azure-class.R
#'
NULL

# avworkspaces ------------------------------------------------------------

.RAWLS_URL <- "https://rawls.dsde-prod.broadinstitute.org"

#' @describeIn avworkspace-methods List workspaces
#'
#' @return `avworkspaces()`: a `tibble` table of available workspaces
#'
#' @importFrom AnVILBase avworkspaces
#' @exportMethod avworkspaces
setMethod("avworkspaces", signature = c(platform = "azure"), definition =
    function(..., platform = cloud_platform()) {
        qrs <- request(.RAWLS_URL) |>
            req_template("/api/workspaces") |>
            req_auth_bearer_token(az_token()) |>
            req_perform()

        AnVILBase::flatten(qrs) |>
            AnVILBase::avworkspaces_clean()
    }
)

# avworkspace_namespace ---------------------------------------------------

.LEONARDO_URL <- "https://leonardo.dsde-prod.broadinstitute.org"

#' @describeIn avworkspace-methods List the workspace namespace
#'
#' @return `avworkspace_namespace()`: a character string of the workspace
#'   namespace
#'
#' @importFrom AnVILBase avworkspace_namespace
#' @exportMethod avworkspace_namespace
setMethod("avworkspace_namespace", signature = c(platform = "azure"),
    definition = function(..., platform = cloud_platform()) {
        qrs <- request(.LEONARDO_URL) |>
            req_template(
                "/api/v2/runtimes/{workspaceid}",
                workspaceid = .avcache$get("workspaceId")
            ) |>
            req_auth_bearer_token(az_token()) |>
            req_perform() |>
            resp_body_json()

        qrs[[1L]][[c("labels", "saturnWorkspaceNamespace")]]
    }
)

# avworkspace_name --------------------------------------------------------

#' @describeIn avworkspace-methods Obtain the workspace name
#'
#' @return `avworkspace_name()`: a character string of the workspace name
#'
#' @importFrom AnVILBase avworkspace_name
#' @exportMethod avworkspace_name
setMethod("avworkspace_name", signature = c(platform = "azure"), definition =
    function(..., platform = cloud_platform()) {
        .avcache$get("workspaceName")
    }
)

# avworkspace -------------------------------------------------------------

#' @describeIn avworkspace-methods Obtain the current workspace namespace and
#'   name combination
#'
#' @return `avworkspace()`: a character string of the workspace namespace and
#'  name combination
#'
#' @importFrom AnVILBase avworkspace
#' @exportMethod avworkspace
setMethod("avworkspace", signature = c(platform = "azure"), definition =
    function(..., platform = cloud_platform()) {
        paste0(avworkspace_namespace(), "/", avworkspace_name())
    }
)

# avworkspace_clone -------------------------------------------------------

#' @describeIn avworkspace-methods Clone a workspace
#'
#' @inheritParams avnotebooks-methods
#' @inheritParams azure-methods
#'
#' @param namespace `character(1)` AnVIL workspace namespace as returned by,
#'   e.g., `avworkspace_namespace()`
#'
#' @param name `character(1)` AnVIL workspace name as returned by, eg.,
#'   `avworkspace_name()`.
#'
#' @param to_namespace `character(1)` workspace (billing account) in
#'     which to make the clone.
#'
#' @param to_name `character(1)` name of the cloned workspace.
#'
#' @param bucket_location `character(1)` region in which bucket
#'   attached to the workspace should be created. The default is set to a
#'   single region ("US"); multi-region is available but more costly.
#'
#' @return `avworkspace_clone()`: called for the side-effect of cloning a
#'  workspace to a new namespace and name.
#'
#' @importFrom AnVILBase avworkspace_clone
#' @exportMethod avworkspace_clone
setMethod("avworkspace_clone", signature = c(platform = "azure"),
    definition = function(
        namespace = avworkspace_namespace(),
        name = avworkspace_name(),
        to_namespace = namespace,
        to_name,
        bucket_location = "US",
        ..., platform = cloud_platform()
    ) {
        stopifnot(
            isScalarCharacter(to_name),
            isScalarCharacter(to_namespace),
            isScalarCharacter(bucket_location),
            `source and destination 'namespace/name' must be different` =
                !identical(namespace, to_namespace) || !identical(name, to_name)
        )

        api_endpoint <-
            "/api/workspaces/{workspaceNamespace}/{workspaceName}/clone"

        request(.RAWLS_URL) |>
            req_template(
                api_endpoint,
                workspaceNamespace = namespace,
                workspaceName = name
            ) |>
            req_auth_bearer_token(az_token()) |>
            req_body_json(
                list(
                    namespace = to_namespace,
                    name = to_name,
                    authorizationDomain = list(), # []
                    attributes = structure(list(), .Names = character()), # {}
                    copyFilesWithPrefix = "analyses/",
                    # createdBy = userName,
                    bucketLocation = bucket_location,
                    enhancedBucketLogging = FALSE
                )
            ) |>
            req_perform()

        paste(to_namespace, to_name, sep = "/")
    }
)

#' @name az-utilities
#'
#' @title az health check helpers
#'
#' @description These functions provide checks for essential workspace tools and
#'   variables. `az_exists` checks for the presence of the `az` command line
#'   utility. `az_healthcheck` checks for the presence of the `az` command line
#'   as well as the essential environment variables.
#'
#' @return az_exists returns a logical value indicating the presence of the `az`
#'   command line utility. `az_healthcheck` returns a logical value indicating
#'   the presence of the `az` command line utility and the essential environment
#'   variables.
#'
#' @examples
#' if (interactive()) {
#'     az_exists()
#'     az_healthcheck()
#' }
#' @export
az_exists <- function() {
    result <- tryCatch({
        .az_find("az")
    }, error = function(...) "")
    nchar(result) > 0L
}

#' @rdname az-utilities
#'
#' @importFrom BiocBaseUtils isScalarCharacter
#' @export
az_health_check <- function() {
    if (!az_exists())
        warning("The 'az' command line utility is not available", call. = FALSE)
    keys <- .avcache$keys()
    nzchars <-  vapply(
        keys, function(x) isScalarCharacter(.avcache$get(x)), logical(1L)
    )
    notfounds <- paste(keys[!nzchars], collapse = ", ")
    if (!nzchar(notfounds))
        warning(
            "The environment variable(s) ", notfounds, " are not set.",
            call. = FALSE
        )
    apps_running <- .apps_running()
    if (is.null(apps_running))
        warning("No apps were found.", call. = FALSE)
    else if (!all(apps_running))
        warning(
            "The following apps are not running: ",
            paste(names(apps_running)[!apps_running], collapse = ", "),
            call. = FALSE
        )
    all(nzchars)
}

#' @name az_copy-helpers
#'
#' @title Azure Copy command line utility helpers
#'
#' @description These functions invoke the `azcopy` command line utility. The
#'   utilities make use of a managed SAS token to mainly transfer files from the
#'   Azure workspace to the Azure Storage container. See `az_sas_token` for
#'   credential details. The results of `azcopy copy` commands are returned as
#'   an `azcopyStatus` object which has S3 methods to print and convert to
#'   logical.
#'
#' @inheritParams azure-methods
#'
#' @details
#' * `az_copy_from_storage` - copy a file from the Azure Storage Container to
#'   the workspace environment
#' * `az_copy_to_storage` - copy a file from the workspace environment to the
#'   Azure Storage Container
#'
#' @param from `character(1)` A relative file path corresponding to
#'   either the remote (`az_copy_from_storage`) or local (`az_copy_to_storage`)
#'   file location. Remote locations should be relative to the base directory in
#'   the Azure Storage Container e.g., `analyses/jupyter.log`.
#'
#' @param to `character(1)` A relative file path corresponding to
#'   either the remote (`az_copy_to_storage`) or local (`az_copy_from_storage`)
#'   file location. Remote locations should be relative to the base directory in
#'   the Azure Storage Container. When not specified, it will default to the
#'   base directory of the remote location. The `to` path can be a folder path
#'   but must end in a forward slash (`/`). If the `to` path points to a
#'   non-existent directory, it will be created.
#'
#' @return
#' * `az_copy_from_storage` - called for the side effect of copying a file
#'   __from__ the Azure Storage Container
#' * `az_copy_to_storage` - called for the side effect of copying a file __to__
#'   the Azure Storage Container
#'
#' @examples
#' if (interactive()) {
#'
#'   ## local -> remote
#'   az_copy_to_storage("jupyter.log", "analyses/jupyter.log")
#'   az_copy_to_storage("jupyter.log", "analyses/test/")
#'
#'   ## placed in the base storage UUID directory
#'   az_copy_to_storage("jupyter.log")
#'
#'   ## remote -> local
#'   az_copy_from_storage("analyses/jupyter.log", "jupyter.log")
#'   ## download to the current directory
#'   az_copy_from_storage("analyses/jupyter.log")
#'
#' }
#' @keywords internal
az_copy_from_storage <-
    function(from, to = "./", recursive = FALSE, dry = TRUE)
{
    stopifnot(
        isScalarCharacter(from), isScalarCharacter(to),
        isScalarLogical(recursive), isScalarLogical(dry)
    )
    if (endsWith(from, "/"))
        stop("Provide a remote file location in the 'from' input")
    .validate_blob(from)

    if (!endsWith(to, "/"))
        stop("Provide a local directory with a forward slash (e.g., './to/')")

    isdir <- file.info(to)[["isdir"]]
    if (is.na(isdir) || !dir.exists(to))
        dir.create(to, recursive = TRUE)
    if (isTRUE(isdir) || endsWith(to, "/"))
        to <- file.path(normalizePath(to), basename(from))
    else
        to <- file.path(normalizePath(dirname(to)), basename(to))

    recurse <- tolower(as.character(recursive))

    sas_cred <- az_sas_token()
    wscu <- .avcache$get("wscu")
    token <- sas_cred[["token"]]
    path <- paste0(wscu, "/", from, "?")
    path <- paste0(path, token)

    results <- .az_copy(
        shQuote(path), shQuote(to), paste0("--recursive=", recurse),
        if (dry) "--dry-run"
    )
    .azcopyStatus(results)
}

#' @rdname az_copy-helpers
az_copy_to_storage <-
    function(from, to, recursive = FALSE, dry = TRUE)
{
    if (!missing(to))
        stopifnot(
            isScalarCharacter(to)
        )

    stopifnot(
        isScalarCharacter(from),
        isScalarLogical(recursive), isScalarLogical(dry)
    )

    recurse <- tolower(as.character(recursive))

    sas_cred <- az_sas_token()
    wscu <- .avcache$get("wscu")
    token <- sas_cred[["token"]]
    path <- sas_cred[["url"]]

    if (!missing(to)) {
        path <- paste0(wscu, "/", to, "?")
        path <- paste0(path, token)
    }

    results <- .az_copy(
        .az_shQuote(from), shQuote(path), paste0("--recursive=", recurse),
        if (dry) "--dry-run"
    )
    .azcopyStatus(results)
}

.azcopyStatus <- function(txtlist) {
    if (!is.list(txtlist) && is.character(txtlist))
        txtlist <- list(txtlist)
    res <- lapply(txtlist, .parse_job_status_text)
    class(res) <- c("azcopyStatus", class(res))
    res
}

.clean_summary <- function(txt) {
    jobsum <- grep("Job.*summary", txt)
    if (!length(txt) || !length(jobsum))
        return(NA_character_)
    jobsumInd <- jobsum + 1
    summary <- txt[jobsumInd:length(txt)]
    strwrap(summary, indent = 2L, exdent = 4L, width = 80)
}

.parse_job_status_text <- function(txt) {
    txt <- sub("\\r", "", txt)
    txt <- Filter(nzchar, txt)
    statusLine <- grep(".*Done.*Failed", txt, value = TRUE)
    status <- grepl("100.0 %", statusLine)
    info <- Filter(function(x) startsWith(x, "INFO"), txt)
    dryrun <- Filter(function(x) startsWith(x, "DRYRUN"), txt)
    if (length(dryrun))
        statusLine <- "100.0 %, 1 DRYRUN"
    summary <- .clean_summary(txt)
    logLine <- grep("Log file is located at:", txt, value = TRUE)
    logfile <- gsub("Log file is located at: (.*)", "\\1", logLine)
    jobLine <- grep("Job (.*) has started", txt, value = TRUE)
    jobId <- gsub("Job (.*) has started", "\\1", jobLine)
    result <- list(
        INFO = if (length(info)) info else NA_character_,
        DRYRUN = if (length(dryrun)) dryrun else NA_character_,
        statusLine = if (length(statusLine)) statusLine else NA_character_,
        Status = if (length(status)) status else FALSE,
        summary = if (length(summary)) summary else NA_character_,
        logFile = if (length(logfile)) logfile else NA_character_,
        jobId = if (length(jobId)) jobId else NA_character_
    )
}

#' @describeIn az_copy-helpers Convert results of azcopy operations to logical
#'   values
#'
#' @param x `azcopyStatus` object to be checked; usually the output of `avcopy`
#'   operations
#'
#' @export
as.logical.azcopyStatus <- function(x, ...) {
    vapply(x, function(x) x$Status, logical(1))
}

#' @describeIn az_copy-helpers Print the results of `azcopy` operations
#'
#' @param ... Additional arguments (not used).
#'
#' @param verbose `logical(1)` Print the `INFO` lines from the `azcopy` output
#'
#' @export
print.azcopyStatus <- function(x, ..., verbose = FALSE) {
    for (i in seq_along(x)) {
        if (verbose)
            cat(
                paste(x[[i]]$INFO, collapse = "\n"), "\n"
            )
        cat(
            "LogFile: ", x[[i]]$logFile, "\n  ",
            x[[i]]$statusLine, "\n",
            sep = ""
        )
    }
}

#' @describeIn az_copy-helpers Get a summary of the results of `azcopy`
#'   operations
#'
#' @param object `azcopyStatus` object to be summarized; usually the output of
#'   `avcopy` operations
#'
#' @export
summary.azcopyStatus <- function(object, ...) {
    for (i in seq_along(object)) {
        cat(
            "Job ID: ", object[[i]]$jobId, "\n",
            paste(object[[i]]$summary, collapse = "\n"), "\n",
            sep = ""
        )
    }
}

# ensure .azcopyStatus works ----------------------------------------------

oneres <- c(
    "INFO: Scanning...", paste0("DRYRUN: copy /home/jupyter/jupyter.log to",
    "https://lza90fd4fb79c636349cdaad.blob.core.windows.net/",
    "sc-995e7d70-320f-42d9-b731-dfb4bfb84708/analyses/jupyter.log")
)

x <- AnVILAz:::.azcopyStatus(oneres)
expect_true(is.list(x))
expect_true(inherits(x, "azcopyStatus"))
expect_identical(length(x), 1L)
expect_identical(
    capture.output(print(x)),
    c("LogFile: NA", "  100.0 %, 1 DRYRUN")
)
expect_identical(
    unlist(x, recursive = FALSE),
    list(
        INFO = oneres[1],
        DRYRUN = oneres[2],
        statusLine = "100.0 %, 1 DRYRUN",
        Status = FALSE,
        summary = NA_character_,
        logFile = NA_character_,
        jobId = NA_character_
    )
)

fullres <- c(
    "INFO: Scanning...",
    paste0(
        "INFO: Any empty folders will not be processed, because source ",
        "and/or destination doesn't have full folder support"
    ),
    "",
    "Job c48205ab-73cf-de44-4669-0eeca2ddf9f9 has started",
    paste0(
        "Log file is located at: /home/jupyter/.azcopy/",
        "c48205ab-73cf-de44-4669-0eeca2ddf9f9.log"
    ),
    "",
    "INFO: azcopy 10.18.1: A newer version 10.25.1 is available to download",
    "",
    paste0(
        "\r100.0 %, 1 Done, 0 Failed, 0 Pending, 0 Skipped, ",
        "1 Total, 2-sec Throughput (Mb/s): 0.0452"
    ),
    "",
    "",
    "Job c48205ab-73cf-de44-4669-0eeca2ddf9f9 summary",
    "Elapsed Time (Minutes): 0.0347",
    "Number of File Transfers: 1",
    "Number of Folder Property Transfers: 0",
    "Number of Symlink Transfers: 0",
    "Total Number of Transfers: 1",
    "Number of File Transfers Completed: 1",
    "Number of Folder Transfers Completed: 0",
    "Number of File Transfers Failed: 0",
    "Number of Folder Transfers Failed: 0",
    "Number of File Transfers Skipped: 0",
    "Number of Folder Transfers Skipped: 0",
    "TotalBytesTransferred: 11762",
    "Final Job Status: Completed",
    ""
)

x <- AnVILAz:::.azcopyStatus(fullres)

expect_true(is.list(x))
expect_true(inherits(x, "azcopyStatus"))
expect_identical(length(x), 1L)
out <- capture.output(print(x))
expect_identical(
   out,
   c(
       "LogFile: /home/jupyter/.azcopy/c48205ab-73cf-de44-4669-0eeca2ddf9f9.log",
       paste0(
            "  100.0 %, 1 Done, 0 Failed, 0 Pending, 0 Skipped, 1 Total, ",
            "2-sec Throughput (Mb/s): 0.0452"
       )
   )
)
expect_identical(
    unlist(x, recursive = FALSE),
    list(
        INFO = grep("INFO", fullres, value = TRUE),
        DRYRUN = NA_character_,
        statusLine = paste0(
            "100.0 %, 1 Done, 0 Failed, 0 Pending, 0 Skipped, 1 Total, ",
            "2-sec Throughput (Mb/s): 0.0452"
        ),
        Status = TRUE,
        summary = c(
            "  Elapsed Time (Minutes): 0.0347",
            "  Number of File Transfers: 1",
            "  Number of Folder Property Transfers: 0",
            "  Number of Symlink Transfers: 0",
            "  Total Number of Transfers: 1",
            "  Number of File Transfers Completed: 1",
            "  Number of Folder Transfers Completed: 0",
            "  Number of File Transfers Failed: 0",
            "  Number of Folder Transfers Failed: 0",
            "  Number of File Transfers Skipped: 0",
            "  Number of Folder Transfers Skipped: 0",
            "  TotalBytesTransferred: 11762",
            "  Final Job Status: Completed"
        ),
        logFile =
            "/home/jupyter/.azcopy/c48205ab-73cf-de44-4669-0eeca2ddf9f9.log",
        jobId = "c48205ab-73cf-de44-4669-0eeca2ddf9f9"
    )
)

---
title: "Convert AnVILAz vignette to jupyter notebook"
output: html_document
---

First convert to jupyter notebook:

```sh
python3 -m jupytext --to notebook vignettes/IntroductionToAnVILAz.Rmd
```

Substitute metadata to use R kernel:

```{r, eval=FALSE}
nb <- readLines("vignettes/IntroductionToAnVILAz.ipynb")
jnb <- jsonlite::fromJSON(nb)
jnb$metadata$kernelspec <-
    list(display_name = "R", language = "R", name = "ir")
jsonlite::write_json(jnb, "vignettes/IntroductionToAnVILAz.ipynb")
```

---
title: "Introduction to the AnVILAz package"
author:
- name: Marcel Ramos
  affiliation: Roswell Park Comprehensive Cancer Center
  email: marcel.ramos@sph.cuny.edu
- name: Martin Morgan
  affiliation: Roswell Park Comprehensive Cancer Center
  email: Martin.Morgan@RoswellPark.org
output:
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    code_folding: show
date: "`r doc_date()`"
package: "`r pkg_ver('AnVILAz')`"
vignette: >
  %\VignetteIndexEntry{Introduction to the AnVILAz package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
az_ok <- AnVILAz::has_avworkspace(strict = TRUE, platform = AnVILAz::azure())
knitr::opts_chunk$set(
    collapse = TRUE,
## Related to https://stat.ethz.ch/pipermail/bioc-devel/2020-April/016656.html
    crop = NULL,
    eval = az_ok
)
options(width = 75)
```

# Installation

The package is not yet available from [Bioconductor](https://bioconductor.org).

Install the development version of the _AnVILAz_ package from GitHub
with

```{r install, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager", repos = "https://cran.r-project.org")
BiocManager::install("Bioconductor/AnVILAz")
```

Once installed, load the package with

```{r library, message = FALSE, eval = TRUE, cache = FALSE}
library(AnVILAz)
```

# File Management

For this tutorial we will refer to the Azure Blob Storage service as ABS.
Within the ABS, we are given access to a Container. For more information,
follow this [link](https://learn.microsoft.com/en-us/azure/storage/blobs/storage-blobs-introduction#containers)
to Microsoft's definition of containers and blobs.

## List Azure Blob Storage Container Files

```{r avlist, eval=az_ok}
avlist()
```

The `avlist` command corresponds to a view of the files in the Blob
container on Azure. They can also be accessed via the
[Microsoft Azure Storage Explorer](https://azure.microsoft.com/en-us/products/storage/storage-explorer).

![Azure Storage Explorer](AzStorageExplorer.png)

## Uploading a file

As an example, we load the internal `mtcars` dataset and save it as an `.Rda`
file with `save`. We can then upload this file to the ABS.

```{r upload_prep, eval=az_ok}
data("mtcars", package = "datasets")
test <- head(mtcars)
save(test, file = "mydata.Rda")
```

Now we can upload the data to the `analyses/` folder in the Azure Blob Storage
(ABS) Container.

```{r avcopy_upload_rda, eval=az_ok}
avcopy("mydata.Rda", "analyses/")
```

We can also use a small log file for demonstration purposes. The `jupyter.log`
file is already present in our workspace directory.

```{r avcopy_upload_log, eval=az_ok}
avcopy("jupyter.log", "analyses/")
```

## Deleting a file

We can remove the data with `avremove` and the _relative_ path to the `.Rda`
file.

```{r avremove, eval=az_ok}
avremove("analyses/mydata.Rda")
```

## Downloading from the ABS

The reverse operation is also possible with a remote and local paths as the
first and second arguments, respectively.

```{r avcopy_download, eval=az_ok}
avcopy("analyses/jupyter.log", "./test/")
```

## Folder-wise upload to ABS

To upload an entire folder, we can use `avbackup`. Note that the entire `test`
folder becomes a subfolder of the remote `analyses` folder in this example.

```{r avbackup, eval=az_ok}
avbackup("./test/", "analyses/")
```

## Folder-wise download from ABS

By default, the entire `source` directory will be copied to the current working
directory `"."`, i.e., the base workspace directory.

```{r avrestore, eval=az_ok}
avrestore("analyses/test")
```

You may also move this to another folder by providing a folder name as the
second argument.

```{r avrestore_dest, eval=az_ok}
avrestore("analyses/test", "test")
```

# The `DATA` tab

## `mtcars` example

First we create an example dataset for uploading to the `DATA` tab. We create a
`model_id` column from the `rownames`.

```{r avtable_prep, eval=az_ok}
library(dplyr)
mtcars_tbl <-
    mtcars |>
    as_tibble(rownames = "model_id") |>
    mutate(model_id = gsub(" ", "-", model_id))
```

## Uploading data

The `avtable_import` command takes an existing R object (usually a `tibble`) and
uploads to the `DATA` tab in the AnVIL User Interface. The `table` argument will
set the name of the table. We also need to provide the `primaryKey` which
corresponds to the column name that uniquely identifies each row in the data.
Typically, the `primaryKey` column provides a list of patient or UUID
identifiers and is in the first column of the data.

```{r avtable_import, eval=az_ok}
mtcars_tbl |> avtable_import(table = "testData", primaryKey = "model_id")
```

## Downloading data

The `avtable` function will pull the data from the `DATA` tab and represent the
data locally as a `tibble`. It works by using the same `type` identifier (i.e.,
the `table` argument) that was used when the data was uploaded.

```{r avtable, eval=az_ok}
model_data <- avtable(table = "testData")
head(model_data)
```

## Delete a row in the table

The API allows deletion of specific rows in the data using
`avtable_delete_values`. To indicate which row to delete, provide the a value or
set of values that correspond to row identifiers in the `primaryKey`. In this
example, we remove the `AMC-Javelin` entry from the data. We are left with 31
records.

```{r avtable_delete_values, eval=az_ok}
avtable_delete_values(table = "testData", values = "AMC-Javelin")
```

## Delete entire table

To remove the entire table from the `DATA` tab, we can use the `avtable_delete`
method with the corresponding table name.

```{r avtable_delete, eval=az_ok}
avtable_delete(table = "testData")
```

# Bug Reports

If you experience issues, please feel free to contact us with a reproducible
example on GitHub:

<https://github.com/Bioconductor/AnVILAz/issues>

# Session information {.unnumbered}

```{r sessionInfo, echo = FALSE, eval = TRUE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/az_copy-helpers.R
\name{az_copy-helpers}
\alias{az_copy-helpers}
\alias{az_copy_from_storage}
\alias{az_copy_to_storage}
\alias{as.logical.azcopyStatus}
\alias{print.azcopyStatus}
\alias{summary.azcopyStatus}
\title{Azure Copy command line utility helpers}
\usage{
az_copy_from_storage(from, to = "./", recursive = FALSE, dry = TRUE)

az_copy_to_storage(from, to, recursive = FALSE, dry = TRUE)

\method{as.logical}{azcopyStatus}(x, ...)

\method{print}{azcopyStatus}(x, ..., verbose = FALSE)

\method{summary}{azcopyStatus}(object, ...)
}
\arguments{
\item{from}{\code{character(1)} A relative file path corresponding to
either the remote (\code{az_copy_from_storage}) or local (\code{az_copy_to_storage})
file location. Remote locations should be relative to the base directory in
the Azure Storage Container e.g., \code{analyses/jupyter.log}.}

\item{to}{\code{character(1)} A relative file path corresponding to
either the remote (\code{az_copy_to_storage}) or local (\code{az_copy_from_storage})
file location. Remote locations should be relative to the base directory in
the Azure Storage Container. When not specified, it will default to the
base directory of the remote location. The \code{to} path can be a folder path
but must end in a forward slash (\code{/}). If the \code{to} path points to a
non-existent directory, it will be created.}

\item{recursive}{\code{logical(1)} Whether to recursively move or remove files in
a directory. Only applies to \code{avremove}, \code{avbackup}, and \code{avrestore}.
Default is \code{TRUE} for \code{backup} and \code{restore} operations and \code{FALSE} for
\code{avremove}.}

\item{dry}{\code{logical(1)} Whether to perform a dry run i.e., add the
\code{--dry-run} flag to the command.}

\item{x}{\code{azcopyStatus} object to be checked; usually the output of \code{avcopy}
operations}

\item{...}{Additional arguments (not used).}

\item{verbose}{\code{logical(1)} Print the \code{INFO} lines from the \code{azcopy} output}

\item{object}{\code{azcopyStatus} object to be summarized; usually the output of
\code{avcopy} operations}
}
\value{
\itemize{
\item \code{az_copy_from_storage} - called for the side effect of copying a file
\strong{from} the Azure Storage Container
\item \code{az_copy_to_storage} - called for the side effect of copying a file \strong{to}
the Azure Storage Container
}
}
\description{
These functions invoke the \code{azcopy} command line utility. The
utilities make use of a managed SAS token to mainly transfer files from the
Azure workspace to the Azure Storage container. See \code{az_sas_token} for
credential details. The results of \verb{azcopy copy} commands are returned as
an \code{azcopyStatus} object which has S3 methods to print and convert to
logical.
}
\details{
\itemize{
\item \code{az_copy_from_storage} - copy a file from the Azure Storage Container to
the workspace environment
\item \code{az_copy_to_storage} - copy a file from the workspace environment to the
Azure Storage Container
}
}
\section{Functions}{
\itemize{
\item \code{as.logical(azcopyStatus)}: Convert results of azcopy operations to logical
values

\item \code{print(azcopyStatus)}: Print the results of \code{azcopy} operations

\item \code{summary(azcopyStatus)}: Get a summary of the results of \code{azcopy}
operations

}}
\examples{
if (interactive()) {

  ## local -> remote
  az_copy_to_storage("jupyter.log", "analyses/jupyter.log")
  az_copy_to_storage("jupyter.log", "analyses/test/")

  ## placed in the base storage UUID directory
  az_copy_to_storage("jupyter.log")

  ## remote -> local
  az_copy_from_storage("analyses/jupyter.log", "jupyter.log")
  ## download to the current directory
  az_copy_from_storage("analyses/jupyter.log")

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avtable-methods.R
\name{avtable-methods}
\alias{avtable-methods}
\alias{avtable}
\alias{avtables}
\alias{avtable_import}
\alias{avtable_import_set}
\alias{avtable_delete}
\alias{avtable_delete_values}
\alias{avtable,azure-method}
\alias{avtables,azure-method}
\alias{avtable_import,azure-method}
\alias{avtable_import_set,azure-method}
\alias{avtable_delete,azure-method}
\alias{avtable_delete_values,azure-method}
\title{AnVIL Azure table ("type") methods}
\usage{
\S4method{avtable}{azure}(table, ..., platform = cloud_platform())

\S4method{avtables}{azure}(..., platform = cloud_platform())

\S4method{avtable_import}{azure}(
  .data,
  table,
  entity = names(.data)[[1L]],
  ...,
  platform = cloud_platform()
)

\S4method{avtable_import_set}{azure}(
  .data,
  origin,
  set = names(.data)[[1]],
  member = names(.data)[[2]],
  ...,
  platform = cloud_platform()
)

\S4method{avtable_delete}{azure}(table, ..., platform = cloud_platform())

\S4method{avtable_delete_values}{azure}(table, values, ..., platform = cloud_platform())
}
\arguments{
\item{table}{\code{character(1)} The name of the table / type}

\item{...}{Additional arguments passed to lower level functions (not used)}

\item{platform}{\code{azure()} The cloud platform class to dispatch on as given by
\link[AnVILBase:cloud_platform]{AnVILBase::cloud_platform}. Typically not set manually as
\code{cloud_platform()} returns the \code{"azure"} class for Azure workspaces on
AnVIL.}

\item{.data}{\code{tibble()} The dataset chiefly from the \code{avtable()} operation}

\item{entity}{The entity name, i.e., the name of the column in the table that
provides the keys for the data (a.k.a. \code{primaryKey}). By default, the first
column in the table. The keys cannot contain special characters or spaces.}

\item{origin}{\code{character(1)} name of the type (entity table) used to create
the set e.g "sample", "participant", etc.}

\item{set}{\code{character(1)} column name of \code{.data} identifying the set(s) to be
created, i.e., the grouping variable.}

\item{member}{\code{character(1)} column name of \code{.data} identifying the member(s)
of the set(s) or groups. The values in this column may repeat if
an ID is in more than one set.}

\item{values}{\code{character()} vector of \code{primaryKey} values corresponding to
rows to be deleted}
}
\value{
\code{avtable}: a \code{tibble()} corresponding to the data with the name as
given by \code{table}

\code{avtables}: a \code{tibble()} with columns \code{table}, \code{count}, and
\code{colnames} corresponding to the tables / types available in the current
workspace

\code{avtable_import()}: called for the side effect of uploading the data
to the DATA tab

\code{avtable_import_set()}: a \code{character(1)} name of the imported
tibble.

\code{avtable_delete}: a \code{logical(1)} indicating success or failure

\code{avtable_delete_values()}: a \code{logical(1)} vector indicating success
or failure for each value in \code{values}
}
\description{
Methods for working with AnVIL Azure tables. These are referred
to as "types" in the AnVIL Workspace Data Service (WDS) API.
}
\details{
\code{avtable_import_set()} creates new rows in a table \verb{<origin>_set}.
One row will be created for each distinct value in the column identified by
\code{set}. Each row entry has a corresponding column \verb{<origin>} linking to one
or more rows in the \verb{<origin>} table, as given in the \code{member} column. The
operation is somewhat like \code{split(member, set)}.
}
\section{Functions}{
\itemize{
\item \code{avtable(azure)}: List the contents of a particular table / type

\item \code{avtables(azure)}: List the available tables / types

\item \code{avtable_import(azure)}: Upload a dataset to the DATA tab

\item \code{avtable_import_set(azure)}: Create a grouping table from an origin dataset

\item \code{avtable_delete(azure)}: Delete a table / type

\item \code{avtable_delete_values(azure)}: Delete rows from a table / type

}}
\examples{
if (interactive()) {
  library(dplyr)
  mtcars_tbl <-
      mtcars |>
      as_tibble(rownames = "model_id") |>
      mutate(model_id = gsub(" ", "-", model_id))

  avtable_import(
    mtcars_tbl,
    table = "testData",
    entity = "model_id"
  )

  avtable("testData")

  avtable("testData") |>                  # new 'testData_set' table
        avtable_import_set("testData", "cyl", "model_id")

  avtable_delete("testData_set")

  avtable_delete_values("testData", "Mazda-RX4")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workspace-dev-ops.R
\name{workspace-dev-ops}
\alias{workspace-dev-ops}
\alias{upload_tsv}
\alias{download_tsv}
\alias{delete_type_id}
\alias{add_type_id}
\alias{get_type_id}
\alias{delete_type}
\title{Functions to work with workspace data for developers}
\usage{
upload_tsv(
  tsv_file,
  type = tools::file_path_sans_ext(basename(tsv_file)),
  primaryKey = NULL
)

download_tsv(type)

delete_type_id(type, id)

add_type_id(row, type, id = row[[1L]])

get_type_id(type, id)

delete_type(type)
}
\arguments{
\item{tsv_file}{\code{character(1)} A path to a tab-separated values file}

\item{type}{\code{character(1)} A nickname for the uploaded dataset important for
retreival. By default, the file name will be used.}

\item{primaryKey}{\code{character(1)} The optional column name to uniquely
identify a record. By default, the first column is used as the primary
key and all values in the column must be unique.}

\item{id}{\code{character(1)} The value in the \code{primaryKey} column that
indicates the row to be removed.}

\item{row}{\code{tibble()} or \code{data.frame()} A single row to add to an existing
table. The row must have the same column names as the table. The
\code{primaryKey} column must be unique.}
}
\value{
\itemize{
\item \code{upload_tsv} - A response list indicating successful upload
\item \code{download_tsv} - A \code{tibble} corresponding to the data labeled with \code{type}
\item \code{delete_type_id}; \code{delete_type} - When successful, a \code{NULL} value
}
}
\description{
These group of functions will allow you to manipulate tables in
the "DATA" tab. Example operations include moving a flat Tab-Separated
Values (TSV) file into the workspace, deleting records, deleting tables,
adding a single row, retrieving a single row, and retrieving the data
table. Note that the API used refers to tables as \code{types}.
}
\details{
These functions use the Workspace Data Services (WDS) API. Current
operations that affect the "DATA" tab include:
\itemize{
\item \code{upload_tsv} - a \code{POST} request using a TSV file that populates the data
\item \code{download_tsv} - a \code{GET} request with the data name (\code{type} argument) in
\code{upload_tsv} to represent the data locally as a \code{tibble}
\item \code{delete_type_id} - a \code{DELETE} request to remove a record or row from
\code{type}
\item \code{add_type_id} - a \code{PUT} request to add a single row to an existing table
(\code{type})
\item \code{get_type_id} - a \code{GET} request to retrieve a single row from an existing
table (\code{type})
\item \code{delete_type} - a \code{DELETE} request to remove then entire data set
(\code{type})
}
}
\examples{
if (interactive()) {
  library(dplyr)
  type <- "model"
  mtcars_tbl <-
      mtcars |>
      as_tibble(rownames = "model_id") |>
      mutate(model_id = gsub(" ", "-", model_id))

  tsv_file <- tempfile()
  readr::write_tsv(mtcars_tbl, tsv_file)
  upload_tsv(
    tsv_file = tsv_file,
    type = "testData",
    primaryKey = "model_id"
  )

  download_tsv("testData")

  ## create an example single row tibble for add_type_id
  datsun <- filter(mtcars_tbl, model_id == "Datsun-710")
  ## change the model_id to be unique
  datsun[["model_id"]] <- "Datsun-512"

  add_type_id(row = datsun, type = "testData")

  get_type_id("testData", "Datsun-512")

  delete_type_id("testData", "Datsun-512")

  delete_type("testData")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AnVILAz-package.R
\docType{package}
\name{AnVILAz-package}
\alias{AnVILAz}
\alias{AnVILAz-package}
\title{AnVILAz: R / Bioconductor Support for the AnVIL Azure Platform}
\description{
The AnVIL is a cloud computing resource developed in part by the National Human Genome Research Institute. The AnVILAz package supports end-users and developers using the AnVIL platform in the Azure cloud. The package provides a programmatic interface to AnVIL resources, including workspaces, notebooks, tables, and workflows. The package also provides utilities for managing resources, including copying files to and from Azure Blob Storage, and creating shared access signatures (SAS) for secure access to Azure resources.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Bioconductor/AnVILAz}
  \item Report bugs at \url{https://github.com/Bioconductor/AnVILAz/issues}
}

}
\author{
\strong{Maintainer}: Marcel Ramos \email{marcel.ramos@sph.cuny.edu} (\href{https://orcid.org/0000-0002-3242-0582}{ORCID})

Authors:
\itemize{
  \item Marcel Ramos \email{marcel.ramos@sph.cuny.edu} (\href{https://orcid.org/0000-0002-3242-0582}{ORCID})
  \item Martin Morgan \email{mtmorgan.bioc@gmail.com} (\href{https://orcid.org/0000-0002-5874-8148}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-methods.R
\name{azure-methods}
\alias{azure-methods}
\alias{avcopy,azure-method}
\alias{avlist,azure-method}
\alias{avremove,azure-method}
\alias{avbackup,azure-method}
\alias{avrestore,azure-method}
\alias{avstorage,azure-method}
\title{A number of methods compatible with the Azure platform class.}
\usage{
\S4method{avcopy}{azure}(source, destination, dry = TRUE, ..., platform = cloud_platform())

\S4method{avlist}{azure}(..., platform = cloud_platform())

\S4method{avremove}{azure}(source, recursive = FALSE, ..., platform = cloud_platform())

\S4method{avbackup}{azure}(
  source,
  destination,
  recursive = TRUE,
  ...,
  platform = cloud_platform()
)

\S4method{avrestore}{azure}(
  source,
  destination,
  recursive = TRUE,
  ...,
  platform = cloud_platform()
)

\S4method{avstorage}{azure}(..., platform = cloud_platform())
}
\arguments{
\item{source}{\code{character(1)} A relative file path corresponding to
either the remote (\code{az_copy_from_storage}) or local (\code{az_copy_to_storage})
file location. Remote locations should be relative to the base directory in
the Azure Storage Container e.g., \code{analyses/jupyter.log}.}

\item{destination}{\code{character(1)} A relative file path corresponding to
either the remote (\code{az_copy_to_storage}) or local (\code{az_copy_from_storage})
file location. Remote locations should be relative to the base directory in
the Azure Storage Container. When not specified, it will default to the
base directory of the remote location. The \code{to} path can be a folder path
but must end in a forward slash (\code{/}). If the \code{to} path points to a
non-existent directory, it will be created.}

\item{dry}{\code{logical(1)} Whether to perform a dry run i.e., add the
\code{--dry-run} flag to the command.}

\item{...}{Additional arguments passed to lower level functions (not used)}

\item{platform}{\code{azure()} The cloud platform class to dispatch on as given by
\link[AnVILBase:cloud_platform]{AnVILBase::cloud_platform}. Typically not set manually as
\code{cloud_platform()} returns the \code{"azure"} class for Azure workspaces on
AnVIL.}

\item{recursive}{\code{logical(1)} Whether to recursively move or remove files in
a directory. Only applies to \code{avremove}, \code{avbackup}, and \code{avrestore}.
Default is \code{TRUE} for \code{backup} and \code{restore} operations and \code{FALSE} for
\code{avremove}.}
}
\value{
\itemize{
\item \code{avlist} - a \code{tibble} of files and metadata
\item \code{avcopy} - called for the side effect of copying a file \strong{to} or \strong{from}
the Azure Storage Container depending on the \code{source} and \code{destination}
inputs
\item \code{avremove} - called for the side effect of removing a file or folder
\item \code{avbackup} - called for the side effect of copying a directory \strong{to}
the Azure Storage Container
\item \code{avrestore} - called for the side effect of copying a directory
\strong{from} the Azure Storage Container
\item \code{avstorage} - a URL string of the Azure Storage Container location
\item \code{avworkspaces} - a tibble of workspaces on AnVIL
\item \code{avtable_import} - a response list indicating successful upload
\item \code{avtable_delete_values} - when successful, a \code{NULL} value
}
}
\description{
A number of methods compatible with the Azure platform class.
}
\details{
The \code{recursive} argument for \code{avbackup} and \code{avrestore} is set to
\code{TRUE} by default and \code{FALSE} for \code{avremove}. Note that wildcards are not
supported for local or remote paths.
}
\section{Functions}{
\itemize{
\item \code{avcopy(azure)}: a generalized interface for either
\code{az_copy_from_storage} or \code{az_copy_to_storage}; deduced from the \code{source}
and \code{destination} inputs

\item \code{avlist(azure)}: list all the files in the Azure Storage Container

\item \code{avremove(azure)}: remove a file or directory from the Azure Storage

\item \code{avbackup(azure)}: copy a directory from the workspace environment to
the Azure Storage Container

\item \code{avrestore(azure)}: copy a file or directory from the Azure Storage
Container to the workspace environment

\item \code{avstorage(azure)}: The base URI string used to move data to and from
the Azure Storage Container

}}
\examples{
if (interactive()) {

  avlist()

  ## local -> remote
  ## using general interface avcopy
  avcopy("jupyter.log", "analyses/jupyter.log")

  ## upload a directory
  avbackup("./test/", "analyses/test/")

  ## using general interface az_copy
  avcopy("analyses/jupyter.log", "./jupyter.log")

  ## download a directory
  avrestore("analyses/test/", "./test/")

  avremove("analyses/jupyter.log")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/authentication.R
\name{az_token}
\alias{az_token}
\title{Generate an Azure authentication token with the \code{az} command line
utility}
\usage{
az_token()
}
\value{
A character string containing the authentication token with a
"Bearer" prefix.
}
\description{
This function generates an Azure authentication token with the
\code{az} command line utility. The token is used to authenticate with the Azure
services. This function is called internally by the \verb{az_*} functions. It
is not meant to be called directly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workspace_env.R
\name{workspace-env}
\alias{workspace-env}
\alias{workspace_id}
\alias{workspace_storage_container_id}
\alias{workspace_storage_container_url}
\alias{workspace_data_service_url}
\alias{workspace_storage_cont_id}
\alias{workspace_storage_cont_url}
\alias{wds_api_version}
\alias{cbas_url}
\title{Access Terra on Azure workspace session variables}
\usage{
workspace_id()

workspace_storage_cont_id()

workspace_storage_cont_url()

wds_api_version()

workspace_data_service_url()

cbas_url()
}
\value{
\itemize{
\item \code{workspace_id} - A UUID string referring to "workspaceId" or "workspaceid"
in API calls
\item \code{workspace_storage_cont_id} - A UUID string identifiying the resource
storage container owned by the user account, a.k.a. "resourceId"
\item \code{workspace_storage_cont_url} - The base URI string used to move data
to and from the Azure Storage Container
\item \code{wds_api_version} - The version of the Workspace Data Service API, defaults
to "v0.2"
\item \code{workspace_data_service_url} - The base URI string used to move data to
to and from the workspace "DATA" tab
\item \code{cbas_url} - The base URI string used to query the workflow submission
history
}
}
\description{
A group of functions that return environment variables in the
Terra Azure workspace. The Workspace Data Service URL sends out an API
\code{GET} request to obtain the data services URL for uploading data to the
workspace "DATA" tab.
}
\examples{
workspace_id()
workspace_storage_cont_id()
workspace_storage_cont_url()
if (interactive()) {
  workspace_data_service_url()
  cbas_url()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_avworkspace-methods.R
\name{has_avworkspace-methods}
\alias{has_avworkspace-methods}
\alias{has_avworkspace,azure-method}
\title{Helper to check if the current environment is within an Azure
workspace}
\usage{
\S4method{has_avworkspace}{azure}(strict = FALSE, ..., platform = cloud_platform())
}
\arguments{
\item{strict}{\code{logical(1)} Whether to include a check for an existing
\code{avworkspace_name()} setting. Default \code{FALSE}.}

\item{...}{Arguments passed to the methods.}

\item{platform}{A Platform derived class indicating the AnVIL environment,
currently, \code{azure} and \code{gcp} classes are compatible.}
}
\value{
\code{logical(1)} \code{TRUE} if the AnVIL environment is set up properly to
interact with Azure, otherwise \code{FALSE}.
}
\description{
\code{has_avworkspace()} checks that the AnVIL environment is set up
to work with Azure. If \code{strict = TRUE}, it also checks that the workspace
name is set.
}
\section{Functions}{
\itemize{
\item \code{has_avworkspace(azure)}: Check if the AnVIL environment is set up

}}
\examples{
has_avworkspace(platform = azure())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/az_sas_token.R
\name{az_sas_token}
\alias{az_sas_token}
\title{Obtain the Shared Access Signature (SAS) for the Azure Storage}
\usage{
az_sas_token(sasExpirationDuration = 28800)
}
\arguments{
\item{sasExpirationDuration}{\code{numeric(1)} The number of seconds until the SAS
token expires (default: 28,800 seconds)}
}
\value{
A list of two elements named \code{token} and \code{url}
}
\description{
The function provides a user delegation SAS token for management of
resources. Mainly used in other functions to move files to and from the
Azure Storage Container
}
\examples{
if (interactive()) {
  sas <- az_sas_token()
  sas[["token"]]
  sas[["url"]]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-class.R
\docType{class}
\name{azure-class}
\alias{azure-class}
\alias{.azure}
\alias{azure}
\title{Azure platform class}
\usage{
azure()
}
\value{
An object of class \code{azure}.
}
\description{
This class represents the Azure platform.
}
\examples{
showClass("azure")

az <- azure()
az
showClass(class(az))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/az-utilities.R
\name{az-utilities}
\alias{az-utilities}
\alias{az_exists}
\alias{az_health_check}
\title{az health check helpers}
\usage{
az_exists()

az_health_check()
}
\value{
az_exists returns a logical value indicating the presence of the \code{az}
command line utility. \code{az_healthcheck} returns a logical value indicating
the presence of the \code{az} command line utility and the essential environment
variables.
}
\description{
These functions provide checks for essential workspace tools and
variables. \code{az_exists} checks for the presence of the \code{az} command line
utility. \code{az_healthcheck} checks for the presence of the \code{az} command line
as well as the essential environment variables.
}
\examples{
if (interactive()) {
    az_exists()
    az_healthcheck()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avworkspace-methods.R
\name{avworkspace-methods}
\alias{avworkspace-methods}
\alias{avworkspaces,azure-method}
\alias{avworkspace_namespace,azure-method}
\alias{avworkspace_name,azure-method}
\alias{avworkspace,azure-method}
\alias{avworkspace_clone,azure-method}
\title{AnVIL Azure Workspace methods}
\usage{
\S4method{avworkspaces}{azure}(..., platform = cloud_platform())

\S4method{avworkspace_namespace}{azure}(..., platform = cloud_platform())

\S4method{avworkspace_name}{azure}(..., platform = cloud_platform())

\S4method{avworkspace}{azure}(..., platform = cloud_platform())

\S4method{avworkspace_clone}{azure}(
  namespace = avworkspace_namespace(),
  name = avworkspace_name(),
  to_namespace = namespace,
  to_name,
  bucket_location = "US",
  ...,
  platform = cloud_platform()
)
}
\arguments{
\item{...}{Additional arguments passed to lower level functions (not used)}

\item{platform}{\code{azure()} The cloud platform class to dispatch on as given by
\link[AnVILBase:cloud_platform]{AnVILBase::cloud_platform}. Typically not set manually as
\code{cloud_platform()} returns the \code{"azure"} class for Azure workspaces on
AnVIL.}

\item{namespace}{\code{character(1)} AnVIL workspace namespace as returned by,
e.g., \code{avworkspace_namespace()}}

\item{name}{\code{character(1)} AnVIL workspace name as returned by, eg.,
\code{avworkspace_name()}.}

\item{to_namespace}{\code{character(1)} workspace (billing account) in
which to make the clone.}

\item{to_name}{\code{character(1)} name of the cloned workspace.}

\item{bucket_location}{\code{character(1)} region in which bucket
attached to the workspace should be created. The default is set to a
single region ("US"); multi-region is available but more costly.}
}
\value{
\code{avworkspaces()}: a \code{tibble} table of available workspaces

\code{avworkspace_namespace()}: a character string of the workspace
namespace

\code{avworkspace_name()}: a character string of the workspace name

\code{avworkspace()}: a character string of the workspace namespace and
name combination

\code{avworkspace_clone()}: called for the side-effect of cloning a
workspace to a new namespace and name.
}
\description{
AnVIL Azure Workspace methods
}
\section{Functions}{
\itemize{
\item \code{avworkspaces(azure)}: List workspaces

\item \code{avworkspace_namespace(azure)}: List the workspace namespace

\item \code{avworkspace_name(azure)}: Obtain the workspace name

\item \code{avworkspace(azure)}: Obtain the current workspace namespace and
name combination

\item \code{avworkspace_clone(azure)}: Clone a workspace

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avnotebooks-methods.R
\name{avnotebooks-methods}
\alias{avnotebooks-methods}
\alias{avnotebooks}
\alias{avnotebooks_localize}
\alias{avnotebooks_delocalize}
\alias{avnotebooks,azure-method}
\alias{avnotebooks_localize,azure-method}
\alias{avnotebooks_delocalize,azure-method}
\title{Azure Notebook Management}
\usage{
\S4method{avnotebooks}{azure}(local = FALSE, ..., platform = cloud_platform())

\S4method{avnotebooks_localize}{azure}(
  destination = "./analyses",
  dry = TRUE,
  ...,
  platform = cloud_platform()
)

\S4method{avnotebooks_delocalize}{azure}(
  source = "./",
  dry = TRUE,
  ...,
  platform = cloud_platform()
)
}
\arguments{
\item{local}{\code{logical(1)} notebooks located on the workspace
(\code{local = FALSE}, default) or runtime / local instance (\code{local = TRUE}).
When \code{local = TRUE}, the notebook path is
\verb{<workspace_data_service_url()>/analyses}.}

\item{...}{Additional arguments passed to lower level functions (not used)}

\item{platform}{\code{azure()} The cloud platform class to dispatch on as given by
\link[AnVILBase:cloud_platform]{AnVILBase::cloud_platform}. Typically not set manually as
\code{cloud_platform()} returns the \code{"azure"} class for Azure workspaces on
AnVIL.}

\item{destination}{\code{character(1)} or \code{missing} file path to the local file
system directory for synchronization. The default location is
\verb{~/<workspace_data_service_url()>/analyses}. Out-of-date local files are
replaced with the workspace version.}

\item{dry}{\code{logical(1)} Whether to perform a dry run i.e., add the
\code{--dry-run} flag to the command.}

\item{source}{\code{character(1)} or \code{missing} file path to the local file system
directory for synchronization. The default location is the home folder.
Out-of-date local files are replaced with the workspace version.}
}
\value{
\code{avnotebooks()} returns a character vector of files located
in the workspace 'analyses/' folder path, or on the local file
system.
}
\description{
\code{avnotebooks()} lists the notebooks in the current workspace.
}
\section{Functions}{
\itemize{
\item \code{avnotebooks(azure)}: List the notebooks in the current workspace

\item \code{avnotebooks_localize(azure)}: Sync notebooks between the Azure Blob Storage
Container and the local runtime

\item \code{avnotebooks_delocalize(azure)}: Sync notebooks between the local runtime and
the Azure Blob Storage Container

}}
\examples{
if (has_avworkspace(strict = TRUE, platform = azure()))
    avnotebooks()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avworkflow-methods.R
\name{avworkflow-methods}
\alias{avworkflow-methods}
\alias{avworkflow_jobs}
\alias{avworkflow_jobs,azure-method}
\alias{avworkflow_jobs_inputs}
\title{Azure Workflow methods}
\usage{
\S4method{avworkflow_jobs}{azure}(..., platform = cloud_platform())

avworkflow_jobs_inputs()
}
\arguments{
\item{...}{Additional arguments passed to lower level functions (not used)}

\item{platform}{\code{azure()} The cloud platform class to dispatch on as given by
\link[AnVILBase:cloud_platform]{AnVILBase::cloud_platform}. Typically not set manually as
\code{cloud_platform()} returns the \code{"azure"} class for Azure workspaces on
AnVIL.}
}
\value{
\code{avworkflow_jobs()} returns a tibble with the status of the
jobs in the current workspace.
}
\description{
\code{avworkflow_jobs()} reports the status of workflow executions
in the current workspace.
}
\details{
The \code{avworkflow_jobs_inputs()} function returns the input parameters
for the workflow jobs as a \code{tibble}.
}
\section{Functions}{
\itemize{
\item \code{avworkflow_jobs(azure)}: List the status of workflow jobs

}}
\examples{
if (has_avworkspace(strict = TRUE, platform = azure()))
    ## from within AnVIL
    avworkflow_jobs()

}
